import numpy as np
import csv
import os
import cv2

class EpisodeMonitor:
    
    def __init__(self, env, logs_dir=None, columns=None, record=False, render_freq=0, num_envs=1):
        self.env = env
        self.logs_dir = logs_dir
        self.record = record
        self.mujoco_step_counter = 0
        self.columns = columns or [
            "episode", "max_reward", "object_name", "texture_name", 
            "grasped_bodies", "steps", "initial_robot_pose", 
            "initial_object_pose", "is_grasping",
            "gripper_current_position", "grasping_object"
        ]
        
        self.is_vector = num_envs > 1
        self.render_freq = render_freq
        
        if self.is_vector:
            self.num_envs = num_envs
            self.global_episode_counter = 0
            self.episode_idx = np.zeros(self.num_envs, dtype=int)
            self.max_reward = np.zeros(self.num_envs)
            self.steps = np.zeros(self.num_envs, dtype=int)
            self.episode_info = [{} for _ in range(self.num_envs)]
            if self.record:
                self.video_writers = [None] * self.num_envs
                self.video_frames = [[] for _ in range(self.num_envs)]
        else:
            self.episode_idx = 0
            self.max_reward = 0
            self.steps = 0
            self.episode_info = {}
            if self.record:
                self.video_writer = None
                self.video_frames = []
        
        self.csv_writer = None
        self.csv_file = None

        os.makedirs(self.logs_dir, exist_ok=True)
        csv_path = os.path.join(self.logs_dir, "log.csv")
        self.csv_file = open(csv_path, 'w', newline='')
        self.csv_writer = csv.DictWriter(self.csv_file, fieldnames=self.columns, delimiter='\t')
        self.csv_writer.writeheader()
        
        if self.record:
            self.video_dir = os.path.join(self.logs_dir, "videos")
            os.makedirs(self.video_dir, exist_ok=True)
        
        self._callbacks_initialized = False
        self._callbacks_paused = False
        self._callbacks_paused = False
    
    def _setup_callbacks(self):
        if self._callbacks_initialized or self.render_freq <= 0:
            return
        
        if self.is_vector:
            # AsyncVectorEnv doesn't provide direct access to individual envs
            # Callbacks cannot be set up for vectorized environments
            pass
        else:
            def callback():
                self._capture_frame_single()
            actual_env = self.env.unwrapped
            actual_env._render_callback = callback
            actual_env._render_freq = self.render_freq
            actual_env._should_render = actual_env.render_mode == 'human'
        
        self._callbacks_initialized = True
    
    def _capture_frame_single(self):
        if self.render_freq <= 0 or self._callbacks_paused:
            return
        
        try:
            actual_env = self.env.unwrapped
            
            if actual_env._should_render:
                actual_env.render()
            
            if self.record:
                frame_ego = actual_env.robot.get_camera_view(actual_env.robot.camera_names[0])
                frame_exo = actual_env.robot.get_camera_view(actual_env.robot.camera_names[1])
                frame = np.concatenate((frame_ego, frame_exo), axis=1)
                frame = self._process_frame(frame)
                self.video_frames.append(frame)
        except Exception as e:
            pass
    
    def _capture_frame_vectorized(self, env_idx):
        if self.render_freq <= 0 or self._callbacks_paused:
            return
        
        try:
            env = self.env.envs[env_idx]
            actual_env = env.unwrapped
            
            if actual_env._should_render:
                actual_env.render()
            
            if self.record:
                frame_ego = actual_env.robot.get_camera_view(actual_env.robot.camera_names[0])
                frame_exo = actual_env.robot.get_camera_view(actual_env.robot.camera_names[1])
                frame = np.concatenate((frame_ego, frame_exo), axis=1)
                frame = self._process_frame(frame)
                self.video_frames[env_idx].append(frame)
        except Exception as e:
            pass
    
    def _process_frame(self, frame):
        if isinstance(frame, np.ndarray):
            if frame.dtype != np.uint8:
                frame = (frame * 255).astype(np.uint8) if frame.max() <= 1.0 else frame.astype(np.uint8)
            if frame.shape[-1] == 3:
                frame = cv2.cvtColor(frame, cv2.COLOR_RGB2BGR)
        return frame
    
    def reset(self, **kwargs):
        self._callbacks_paused = True
        
        if self.is_vector:
            options = kwargs.get('options', {})
            reset_mask = options.get('reset_mask', np.ones(self.num_envs, dtype=bool))
            
            obs, info = self.env.reset(**kwargs)
            self._setup_callbacks()
            
            for i in np.where(reset_mask)[0]:
                if self.steps[i] == 0 and self.record:
                    self.video_frames[i] = []
                self.max_reward[i] = 0
                self.steps[i] = 0
                self.episode_info[i] = {}
            
            self._callbacks_paused = False
        else:
            if self.steps > 0:
                if self.csv_writer:
                    self._log_episode()
                if self.record:
                    self._save_video()
                self.episode_idx += 1
            obs, info = self.env.reset(**kwargs)
            self._setup_callbacks()
            
            self.max_reward = 0
            self.steps = 0
            self.episode_info = {}
            if self.record:
                self.video_frames = []
            self._callbacks_paused = False
        return obs, info
    
    def step(self, action):
        obs, reward, terminated, truncated, info = self.env.step(action)
        
        if self.render_freq == 0:
            if self.is_vector:
                # AsyncVectorEnv doesn't provide direct access to individual envs
                # Recording from observations if available
                if self.record and 'rgb_ego' in obs and 'rgb_exo' in obs:
                    for i in range(self.num_envs):
                        frame_ego = obs['rgb_ego'][i]
                        frame_exo = obs['rgb_exo'][i]
                        frame = np.concatenate((frame_ego, frame_exo), axis=1)
                        frame = self._process_frame(frame)
                        self.video_frames[i].append(frame)
            else:
                actual_env = self.env.unwrapped
                if actual_env._should_render:
                    actual_env.render()
                if self.record and 'rgb_ego' in obs and 'rgb_exo' in obs:
                    frame_ego = obs['rgb_ego']
                    frame_exo = obs['rgb_exo']
                    frame = np.concatenate((frame_ego, frame_exo), axis=1)
                    frame = self._process_frame(frame)
                    self.video_frames.append(frame)
        
        if self.is_vector:
            self.max_reward = np.maximum(self.max_reward, reward)
            self.steps += 1
            for i in range(self.num_envs):
                if isinstance(info, dict):
                    env_info = {}
                    for k, v in info.items():
                        if k.startswith('_'):
                            continue
                        if isinstance(v, (list, np.ndarray)) and len(v) > i:
                            env_info[k] = v[i]
                        else:
                            env_info[k] = v
                else:
                    env_info = info[i] if isinstance(info, (list, tuple)) else {}
                self.episode_info[i].update(env_info)
            
            dones = np.logical_or(terminated, truncated)
            if isinstance(dones, bool):
                dones = np.array([dones] * self.num_envs)
        else:
            self.max_reward = max(self.max_reward, reward)
            self.steps += 1
            self.episode_info.update(info)
            info['episode_stats'] = {'episode': self.episode_idx, 'max_reward': self.max_reward, 'steps': self.steps}
        
        return obs, reward, terminated, truncated, info
    
    def _log_episode(self, env_idx=None):
        if env_idx is not None:
            log_data = {"episode": self.global_episode_counter, "max_reward": float(self.max_reward[env_idx]), "steps": int(self.steps[env_idx]), **self.episode_info[env_idx]}
            self.global_episode_counter += 1
        else:
            log_data = {"episode": self.episode_idx, "max_reward": self.max_reward, "steps": self.steps, **self.episode_info}
        
        filtered_data = {}
        for col in self.columns:
            value = log_data.get(col, "")
            if isinstance(value, np.ndarray):
                value = value.flatten().tolist()
            filtered_data[col] = value
        
        self.csv_writer.writerow(filtered_data)
        self.csv_file.flush()
    
    def log_episodes(self, env_indices):
        if self.is_vector:
            for i in env_indices:
                if self.csv_writer:
                    self._log_episode(i)
                if self.record:
                    self._save_video(i)
                    self.video_frames[i] = []
                self.episode_idx[i] += 1
    
    def _save_video(self, env_idx=None):
        if env_idx is not None:
            frames = self.video_frames[env_idx]
            episode_num = int(self.episode_idx[env_idx])
            video_filename = f"env_{env_idx}_episode_{episode_num}.mp4"
        else:
            frames = self.video_frames
            episode_num = self.episode_idx
            video_filename = f"episode_{episode_num}.mp4"
        
        if not frames:
            return
        
        video_path = os.path.join(self.video_dir, video_filename)
        height, width = frames[0].shape[:2]
        fourcc = cv2.VideoWriter_fourcc(*'mp4v')
        out = cv2.VideoWriter(video_path, fourcc, 30, (width, height))
        
        for frame in frames:
            out.write(frame)
        
        out.release()
    
    def close(self):
        if self.is_vector:
            for i in range(self.num_envs):
                if self.steps[i] > 0:
                    if self.csv_writer:
                        self._log_episode(i)
                    if self.record:
                        self._save_video(i)
        else:
            if self.steps > 0:
                if self.csv_writer:
                    self._log_episode()
                if self.record:
                    self._save_video()
        if self.csv_file:
            self.csv_file.close()
        self.env.close()
    
    def __getattr__(self, name):
        return getattr(self.env, name)



import os
import numpy as np
import time
import io
import base64
import requests
from PIL import Image
from egogym.utils import pixel_to_world

MOONDREAM_API_URL = "https://api.moondream.ai/v1/point"


class UnprivilegedMoondream:
    
    def __init__(self, env, api_key=None):
        self.env = env
        self.unprivileged_T_world_object = None
        
        if api_key is None:
            api_key = os.environ.get("MOONDREAM_API_KEY")
        
        self.api_key = api_key
    
    def _call_api_with_retry(self, image_base64: str, object_name: str, max_retries=3, base_delay=2):
        """Call Moondream API with retry logic for transient errors."""
        headers = {
            "Content-Type": "application/json",
            "X-Moondream-Auth": self.api_key
        }
        
        payload = {
            "image_url": f"data:image/jpeg;base64,{image_base64}",
            "object": object_name
        }
        
        for attempt in range(max_retries):
            try:
                response = requests.post(MOONDREAM_API_URL, headers=headers, json=payload, timeout=30)
                
                if response.status_code == 200:
                    return response.json()
                elif response.status_code in [503, 429] and attempt < max_retries - 1:
                    delay = base_delay * (2 ** attempt)
                    print(f"API error {response.status_code}: {response.text}. Retrying in {delay}s... (attempt {attempt + 1}/{max_retries})")
                    time.sleep(delay)
                    continue
                else:
                    response.raise_for_status()
            except requests.exceptions.RequestException as e:
                if attempt < max_retries - 1:
                    delay = base_delay * (2 ** attempt)
                    print(f"Request error: {e}. Retrying in {delay}s... (attempt {attempt + 1}/{max_retries})")
                    time.sleep(delay)
                    continue
                else:
                    raise
        
        return None
    
    def _infer_point_sync(self, rgb: np.ndarray, object_name: str) -> np.ndarray:
        """Use Moondream API to locate object in image and return normalized coordinates."""
        if rgb.dtype == np.float32 or rgb.dtype == np.float64:
            rgb = (rgb * 255).astype(np.uint8)
        image = Image.fromarray(rgb)
        
        try:
            buf = io.BytesIO()
            image.save(buf, format='JPEG')
            image_bytes = buf.getvalue()
            image_base64 = base64.b64encode(image_bytes).decode('utf-8')
            
            response = self._call_api_with_retry(image_base64, object_name)
            
            if response and "points" in response and len(response["points"]) > 0:
                point = response["points"][0]
                x_norm = float(point["x"])
                y_norm = float(point["y"])
                
                x_norm = max(0.0, min(1.0, x_norm))
                y_norm = max(0.0, min(1.0, y_norm))
                
                return np.array([x_norm, y_norm], dtype=np.float32)
            
            return np.array([0.5, 0.5], dtype=np.float32)
            
        except Exception as e:
            print(f"Moondream API error: {e}")
            return np.array([0.5, 0.5], dtype=np.float32)

    def reset(self, **kwargs):
        obs, info = self.env.reset(**kwargs)
        
        robot = self.env.unwrapped.robot
        robot.rgb_renderers[robot.camera_names[0]].enable_depth_rendering()
        robot.rgb_renderers[robot.camera_names[0]].update_scene(robot.data, robot.camera_names[0])
        depth = robot.rgb_renderers[robot.camera_names[0]].render()
        robot.rgb_renderers[robot.camera_names[0]].disable_depth_rendering()

        point_norm = self._infer_point_sync(obs["rgb_ego"], obs["object_name"])
        x_norm, y_norm = point_norm
        x = int(x_norm * self.env.unwrapped.render_width)
        y = int(y_norm * self.env.unwrapped.render_height)
        depth_value = depth[y, x] + 0.03
        self.unprivileged_T_world_object = pixel_to_world(x, y, depth_value, self.env.unwrapped.model, self.env.unwrapped.data, robot.camera_names[0], self.env.unwrapped.render_width, self.env.unwrapped.render_height)

        return obs, info
    
    def step(self, action):
        obs, reward, terminated, truncated, info = self.env.step(action)
        
        object_pose = obs["object_pose"].reshape(4,4)
        object_pose[:3, 3] = self.unprivileged_T_world_object
        obs["object_pose"] = object_pose.flatten()

        return obs, reward, terminated, truncated, info
    
    def close(self):
        self.env.close()
    
    def __getattr__(self, name):
        return getattr(self.env, name)



import os
import cv2
import mujoco
from gymnasium.spaces import Box
import numpy as np
from scipy.spatial.transform import Rotation as R

from egogym.egogym import Egogym
from egogym.utils import include_in_scene, position_sampler, make_objects_manager
import egogym.assets.constants as constants

class OpenTask(Egogym):

    def __init__(self, robot="rum", action_space="delta", render_mode=None, render_size=(960,720), num_objs=1, seed=None, objects_set=None):
        super().__init__(robot=robot, action_space=action_space, render_mode=render_mode, render_size=render_size, seed=seed)
        self.num_objs = num_objs
        if objects_set is not None:
            self.objects_manager = make_objects_manager(objects_set, self.np_random, shuffle=False)
        else:
            self.objects_manager = make_objects_manager(constants.all_open_objects_set, self.np_random)
        self.observation_space["handle_pose"] = Box(low=-np.inf, high=np.inf, shape=(16,), dtype=np.float32)

    def make_task_scene(self):
        with open(f"{os.path.dirname(__file__)}/../assets/scenes/open.xml", "r") as f:
            scene_xml = f.read()

        self.object = self.objects_manager.sample()
        scene_xml = self.object.add_to_scene_xml(scene_xml)
        scene_xml = include_in_scene(scene_xml, f"{os.path.dirname(__file__)}/../assets/embodiments/{self.robot.name}/model_open.xml")
        return scene_xml

    def get_obs(self):
        obs = self.get_base_obs()
        obs["handle_pose"] = self.object.get_handle_pose(self.data).reshape(16).astype(np.float32)
        return obs

    def compute_reward(self) -> float:
        perecentage_opened = self.object.get_perecenttage_opened(self.model, self.data)
        return perecentage_opened

    def reset(self, seed=None, options=None):
        super().reset(seed=seed, options=options)
        scene_xml_string = self.make_task_scene()
        self.grasped_bodies = set()
        self.grasping_object = False
        
        gripper_init_pose = np.eye(4)
        gripper_init_pose[:3, :3] = R.from_euler("xyz", np.array([1.5, 0.000, 0.000])).as_matrix()
        self.setup_mujoco(scene_xml_string)
        mujoco.mj_step(self.model, self.data)
        handle_pose = self.object.get_handle_pose(self.data)
        camera_pos = handle_pose[0:3, 3].copy()
        camera_pos += np.array([0.0, -1.0, 0.28])
        self.model.cam_pos[mujoco.mj_name2id(self.model, mujoco.mjtObj.mjOBJ_CAMERA, "exocentric")] = camera_pos
        mujoco.mj_step(self.model, self.data)
        gripper_init_pose[:3, 3] = handle_pose[:3, 3] + np.array([0.0, -0.6, 0.05])
        self.robot.prepare(gripper_init_pose, self.env_step)
        self.env_step()
        self.initial_robot_pose = self.robot.get_camera_pose()

        self.settle_env()
        self.enable_sleeping_islands()

        observation = self.get_obs()
        info = {"object_name": self.object.name}

        return observation, info

    def step(self, action):
        obs, reward, terminated, truncated, info = super().step(action)
        self.grasped_bodies.update(self.robot.get_grasped_bodies())
        grasped_bodies_list = list(self.grasped_bodies)
        if f"{self.object.name}_object" in grasped_bodies_list:
            self.grasping_object = True
        info = {
            "grasped_bodies": grasped_bodies_list,
            "object_name": self.object.name,
            "initial_robot_pose": self.initial_robot_pose,
            "initial_object_pose": self.object.last_set_pose,
            "is_grasping": self.robot.get_grasp(),
            "gripper_current_position": self.robot.get_tcp_pose(),
            "grasping_object": self.grasping_object,
            **info
        }
        return obs, reward, terminated, truncated, info



import pathlib
import warnings
from abc import ABC, abstractmethod
from typing import Union

import torch
import torch.nn as nn

ENCODER_LOADING_ERROR_MSG = (
    "Could not load encoder weights: defaulting to pretrained weights"
)


class AbstractEncoder(nn.Module, ABC):
    def __init__(self):
        super().__init__()

    @property
    @abstractmethod
    def feature_dim(self):
        pass

    def transform(self, x):
        return x

    def to(self, device):
        self.device = device
        return super().to(device)

    @abstractmethod
    def forward(self, x):
        pass

    def load_weights(
        self, weight_path: Union[pathlib.Path, str], strict: bool = True
    ) -> None:
        try:
            checkpoint = torch.load(weight_path, map_location="cpu")
            weight_dict = checkpoint.get("model", checkpoint)

            self.model.load_state_dict(weight_dict, strict=strict)
            return

        except RuntimeError:
            warnings.warn("Couldn't load full model weights. Trying fallback options...")

        fallback_prefixes = [
            "model.",
            "encoder.",
            "module.base_encoder._orig_mod.",
        ]

        for prefix in fallback_prefixes:
            try:
                filtered_weights = {
                    k.replace(prefix, ""): v
                    for k, v in weight_dict.items()
                    if k.startswith(prefix)
                }

                filtered_weights.pop("fc.weight", None)
                filtered_weights.pop("fc.bias", None)

                self.model.load_state_dict(filtered_weights, strict=strict)
                return

            except RuntimeError:
                warnings.warn(f"Failed loading with prefix: '{prefix}'")

        warnings.warn(ENCODER_LOADING_ERROR_MSG)



import os
import numpy as np
import pandas as pd
import altair as alt
from scipy.stats import beta

# Register custom font for PNG export
alt.themes.register('custom_theme', lambda: {
    'config': {
        'title': {'font': 'Produkt'},
        'axis': {'labelFont': 'Produkt', 'titleFont': 'Produkt'},
        'legend': {'labelFont': 'Produkt', 'titleFont': 'Produkt'},
        'mark': {'font': 'Produkt'},
        'text': {'font': 'Produkt'},
    }
})
alt.themes.enable('custom_theme')

BASE_DIR = "logs"
REWARD_THRESHOLD = 0.03

def compute_success_from_csv(csv_path):
    df = pd.read_csv(csv_path, sep="\t")
    successes = (df["max_reward"] > REWARD_THRESHOLD).sum()
    total = len(df)
    return successes, total


def plot_pi0_bars():
    model_folder = "pi0"
    num_objects = [1, 2, 3, 4, 5]
    
    rows = []
    
    for n_obj in num_objects:
        possible_folders = [
            f"{n_obj}_objects",
            f"{n_obj}_object",
            f"{n_obj}-objects",
            f"{n_obj}-object",
        ]
        
        csv_path = None
        for folder in possible_folders:
            # First try direct path
            candidate = os.path.join(BASE_DIR, model_folder, folder, "log.csv")
            if os.path.exists(candidate):
                csv_path = candidate
                break
            
            # Try nested evaluation folder structure
            folder_path = os.path.join(BASE_DIR, model_folder, folder)
            if os.path.exists(folder_path) and os.path.isdir(folder_path):
                for subdir in os.listdir(folder_path):
                    subdir_path = os.path.join(folder_path, subdir)
                    if os.path.isdir(subdir_path):
                        candidate = os.path.join(subdir_path, "log.csv")
                        if os.path.exists(candidate):
                            csv_path = candidate
                            break
                if csv_path:
                    break
        
        if csv_path is None:
            print(f"Missing data: π-0.5, {n_obj} objects")
            continue
        
        s, t = compute_success_from_csv(csv_path)
        
        # Beta posterior
        a, b = 1 + s, 1 + (t - s)
        mean = 100 * a / (a + b)
        lo = 100 * beta.ppf(0.025, a, b)
        hi = 100 * beta.ppf(0.975, a, b)
        
        print(f"π-0.5 | {n_obj} objects: {s}/{t} = {mean:.1f}%")
        
        rows.append({
            "num_objects": n_obj,
            "mean": mean,
            "lo": lo,
            "hi": hi,
            "successes": s,
            "total": t
        })
    
    df = pd.DataFrame(rows)
    
    # Create bar chart
    bars = alt.Chart(df).mark_bar(
        color='#E0BE16',
        width=60
    ).encode(
        x=alt.X(
            'num_objects:O',
            title='Number of Objects',
            axis=alt.Axis(
                labelFontSize=20,
                titleFontSize=24,
                labelAngle=0,
                titlePadding=15
            )
        ),
        y=alt.Y(
            'mean:Q',
            title='Success Rate (%)',
            scale=alt.Scale(domain=[0, 100]),
            axis=alt.Axis(
                labelFontSize=20,
                titleFontSize=24,
                titlePadding=14,
                grid=True,
                gridOpacity=0.3
            )
        ),
        tooltip=[
            alt.Tooltip('num_objects:O', title='Objects'),
            alt.Tooltip('mean:Q', title='Success Rate (%)', format='.1f'),
            alt.Tooltip('successes:Q', title='Successes'),
            alt.Tooltip('total:Q', title='Total')
        ]
    )
    
    # Create error bars for 95% CI
    error_bars = alt.Chart(df).mark_errorbar(
        ticks=True,
        thickness=2
    ).encode(
        x=alt.X('num_objects:O'),
        y=alt.Y('lo:Q', title=''),
        y2=alt.Y2('hi:Q')
    )
    
    # Add text labels on top of bars
    text = alt.Chart(df).mark_text(
        dy=-10,
        fontSize=16,
        fontWeight='bold'
    ).encode(
        x=alt.X('num_objects:O'),
        y=alt.Y('mean:Q'),
        text=alt.Text('mean:Q', format='.1f')
    )
    
    # Combine layers
    chart = (bars + error_bars + text).properties(
        width=500,
        height=400,
        title={
            'text': '           π-0.5 Success Rate by Number of Objects',
            'fontSize': 24,
            'anchor': 'start',
            'dx': 40,
            'dy': -10
        },
        padding={'left': 10, 'right': 10, 'top': 40, 'bottom': 40}
    ).configure_view(
        strokeWidth=0
    )
    
    return chart


if __name__ == "__main__":
    chart = plot_pi0_bars()
    chart.save("pi0_bars.html")
    chart.save("pi0_bars.png", scale_factor=3)
    chart.save("pi0_bars.pdf", scale_factor=3)
    print("\nPlot saved to: pi0_bars.html, pi0_bars.png, and pi0_bars.pdf")



all_close_objects_set = [
    "cabinet",
    "drawer"
]

all_open_objects_set = [
    "cabinet",
    "drawer"
]

cabinet_objects_set = [
    "cabinet"
]

drawer_objects_set = [
    "drawer"
]

lite_pick_objects_set = [
    "bottled_drink_2",
    "cup_2",
    "egg_0",
    "hot_dog_0",
    "ketchup_2",
    "kiwi_0",
    "lemon_0",
    "mango_5",
    "orange_2",
    "potato_2",
    "tangerine_3",
    "tomato_1",
    "wine_4",
    "apple_3"
]

diverse_pick_objects_set = [
    "apple_6",
    "tomato_7",
    "can_16",
    "onion_1",
    "bottled_drink_0",
    "kettle_14",
    "spoon_1",
    "mango_19",
    "potato_2",
    "lime_1",
    "can_4",
    "apple_10",
    "water_bottle_4",
    "egg_1",
    "potato_5",
    "egg_4",
    "cupcake_17",
    "hot_dog_3",
    "can_6",
    "mango_11",
    "mushroom_3",
    "bagel_0",
    "candle_2",
    "rolling_pin_0",
    "croissant_2",
    "mug_7",
    "pear_2",
    "kettle_26",
    "bagel_6",
    "bar_3",
    "milk_5",
    "spoon_16",
    "squash_7",
    "sponge_1",
    "eggplant_0",
    "cupcake_3",
    "cereal_12",
    "bowl_19",
    "bread_8",
    "mug_18",
    "donut_0",
    "kettle_11",
    "pan_13",
    "knife_1",
    "bread_13",
    "plate_13",
    "mug_0",
    "bowl_3",
]

plausible_categories = [
    "alcohol",
    "apple",
    "avocado",
    "bagel",
    "baggedfood",
    "baguette",
    "banana",
    "bar",
    "barsoap",
    "beer",
    "bellpepper",
    "bottleddrink",
    "bottledwater",
    "bowl",
    "boxeddrink",
    "boxedfood",
    "bread",
    "broccoli",
    "cake",
    "can",
    "candle",
    "cannedfood",
    "carrot",
    "cereal",
    "cheese",
    "chips",
    "coffeecup",
    "condiment",
    "corn",
    "croissant",
    "cup",
    "cupcake",
    "cucumber",
    "donut",
    "egg",
    "eggplant",
    "fish",
    "fork",
    "garlic",
    "hotdog",
    "jam",
    "jug",
    "ketchup",
    "kettle",
    "kiwi",
    "knife",
    "ladle",
    "lemon",
    "lime",
    "mango",
    "milk",
    "mug",
    "mushroom",
    "onion",
    "orange",
    "pan",
    "peach",
    "pear",
    "potato",
    "rollingpin",
    "scissors",
    "shaker",
    "soapdispenser",
    "spatula",
    "sponge",
    "spoon",
    "spray",
    "squash",
    "steak",
    "sweetpotato",
    "tangerine",
    "teapot",
    "tomato",
    "waffle",
    "waterbottle",
    "wine",
    "yogurt"
]


all_categories_names = [
   "alcohol",
    "apple",
    "avocado",
    "bagel",
    "baggedfood",
    "baguette",
    "banana",
    "bar",
    "barsoap",
    "beer",
    "bellpepper",
    "bottleddrink",
    "bottledwater",
    "bowl",
    "boxeddrink",
    "boxedfood",
    "bread",
    "broccoli",
    "cake",
    "can",
    "candle",
    "cannedfood",
    "carrot",
    "cereal",
    "cheese",
    "chips",
    "coffeecup",
    "condiment",
    "corn",
    "croissant",
    "cup",
    "cucumber", 
    "cupcake",
    "cucumber",
    "donut",
    "egg",
    "eggplant",
    "fish",
    "fork",
    "garlic",
    "hotdog",
    "jam",
    "jug",
    "ketchup",
    "kettle",
    "kiwi",
    "knife",
    "ladle",
    "lemon",
    "lime",
    "mango",
    "milk",
    "mug",
    "mushroom",
    "onion",
    "orange",
    "pan",
    "peach",
    "pear",
    "potato",
    "rollingpin",
    "scissors",
    "shaker",
    "soapdispenser",
    "spatula",
    "sponge",
    "spoon",
    "spray",
    "squash",
    "steak",
    "sweetpotato",
    "tangerine",
    "teapot",
    "tomato",
    "waffle",
    "waterbottle",
    "wine",
    "yogurt"
]

full_eval_objects_set = [
    "alcohol_0",
    "alcohol_1",
    "alcohol_2",
    "alcohol_4",
    "alcohol_5",
    "alcohol_7",
    "alcohol_9",
    "alcohol_10",
    "alcohol_11",
    "alcohol_12",
    "apple_0",
    "apple_1",
    "apple_2",
    "apple_3",
    "apple_4",
    "apple_5",
    "apple_6",
    "apple_7",
    "apple_8",
    "apple_9",
    "avocado_0",
    "avocado_1",
    "avocado_2",
    "avocado_4",
    "avocado_5",
    "avocado_6",
    "avocado_8",
    "bagel_0",
    "bagel_1",
    "bagel_2",
    "bagel_3",
    "bagel_4",
    "bagel_5",
    "bagel_6",
    "bagel_7",
    "bagel_9",
    "bagged_food_0",
    "bagged_food_3",
    "bagged_food_4",
    "bagged_food_5",
    "bagged_food_7",
    "bagged_food_8",
    "bagged_food_9",
    "bagged_food_11",
    "bagged_food_14",
    "baguette_0",
    "baguette_1",
    "baguette_2",
    "baguette_4",
    "baguette_5",
    "baguette_6",
    "banana_1",
    "banana_2",
    "banana_3",
    "banana_4",
    "banana_5",
    "banana_6",
    "banana_7",
    "banana_8",
    "banana_9",
    "banana_10",
    "bar_0",
    "bar_2",
    "bar_3",
    "bar_4",
    "bar_6",
    "bar_7",
    "bar_8",
    "bar_9",
    "bar_10",
    "bar_11",
    "bar_soap_0",
    "bar_soap_1",
    "bar_soap_3",
    "bar_soap_4",
    "beer_0",
    "beer_1",
    "beer_2",
    "beer_3",
    "beer_5",
    "beer_6",
    "beer_7",
    "beer_8",
    "beer_9",
    "beer_10",
    "bell_pepper_0",
    "bell_pepper_1",
    "bell_pepper_2",
    "bell_pepper_3",
    "bottled_drink_0",
    "bottled_drink_1",
    "bottled_drink_2",
    "bottled_drink_4",
    "bottled_drink_5",
    "bottled_water_1",
    "bottled_water_2",
    "bottled_water_3",
    "bottled_water_4",
    "bottled_water_6",
    "bottled_water_7",
    "bottled_water_8",
    "bottled_water_10",
    "bottled_water_11",
    "bottled_water_12",
    "bowl_0",
    "bowl_1",
    "bowl_2",
    "bowl_3",
    "bowl_4",
    "bowl_5",
    "bowl_6",
    "bowl_7",
    "bowl_8",
    "bowl_9",
    "boxed_drink_0",
    "boxed_drink_1",
    "boxed_drink_2",
    "boxed_drink_3",
    "boxed_drink_4",
    "boxed_drink_5",
    "boxed_drink_7",
    "boxed_food_0",
    "boxed_food_1",
    "boxed_food_2",
    "boxed_food_4",
    "boxed_food_6",
    "boxed_food_7",
    "boxed_food_8",
    "boxed_food_9",
    "boxed_food_10",
    "boxed_food_11",
    "bread_0",
    "bread_1",
    "bread_2",
    "bread_3",
    "bread_4",
    "bread_5",
    "bread_6",
    "bread_7",
    "bread_8",
    "bread_9",
    "broccoli_0",
    "broccoli_1",
    "broccoli_3",
    "cake_0",
    "cake_1",
    "cake_3",
    "cake_4",
    "cake_5",
    "cake_6",
    "cake_7",
    "cake_8",
    "cake_9",
    "can_0",
    "can_2",
    "can_3",
    "can_4",
    "can_6",
    "can_8",
    "can_9",
    "can_11",
    "can_12",
    "can_13",
    "candle_0",
    "candle_1",
    "candle_2",
    "candle_3",
    "candle_4",
    "candle_5",
    "candle_6",
    "candle_7",
    "candle_8",
    "candle_9",
    "canned_food_0",
    "canned_food_1",
    "canned_food_3",
    "canned_food_4",
    "canned_food_5",
    "canned_food_6",
    "canned_food_8",
    "canned_food_9",
    "canned_food_10",
    "canned_food_11",
    "carrot_0",
    "carrot_1",
    "carrot_2",
    "carrot_3",
    "carrot_4",
    "carrot_6",
    "carrot_7",
    "cereal_0",
    "cereal_1",
    "cereal_2",
    "cereal_3",
    "cereal_4",
    "cereal_5",
    "cereal_6",
    "cereal_7",
    "cereal_8",
    "cereal_9",
    "cheese_0",
    "cheese_1",
    "cheese_2",
    "cheese_3",
    "cheese_4",
    "cheese_5",
    "cheese_6",
    "cheese_7",
    "cheese_8",
    "chips_0",
    "chips_1",
    "chips_2",
    "chips_3",
    "chips_4",
    "chips_5",
    "chips_6",
    "chips_7",
    "chips_8",
    "chips_9",
    "coffee_cup_1",
    "coffee_cup_2",
    "coffee_cup_3",
    "coffee_cup_4",
    "coffee_cup_6",
    "coffee_cup_7",
    "coffee_cup_8",
    "coffee_cup_9",
    "coffee_cup_10",
    "coffee_cup_11",
    "condiment_0",
    "condiment_2",
    "condiment_3",
    "corn_0",
    "corn_1",
    "corn_3",
    "corn_4",
    "croissant_0",
    "croissant_1",
    "croissant_2",
    "croissant_3",
    "croissant_4",
    "croissant_5",
    "croissant_6",
    "croissant_8",
    "croissant_9",
    "croissant_10",
    "cucumber_0",
    "cucumber_1",
    "cucumber_2",
    "cucumber_3",
    "cucumber_4",
    "cucumber_5",
    "cup_2",
    "cup_3",
    "cup_4",
    "cup_5",
    "cupcake_2",
    "cupcake_3",
    "cupcake_4",
    "cupcake_5",
    "cupcake_6",
    "cupcake_7",
    "cupcake_9",
    "cupcake_11",
    "cupcake_12",
    "cupcake_14",
    "cutting_board_0",
    "cutting_board_1",
    "cutting_board_2",
    "cutting_board_4",
    "cutting_board_5",
    "cutting_board_7",
    "cutting_board_9",
    "cutting_board_11",
    "cutting_board_12",
    "cutting_board_13",
    "donut_0",
    "donut_1",
    "donut_2",
    "donut_3",
    "donut_4",
    "donut_5",
    "donut_6",
    "donut_7",
    "donut_8",
    "donut_9",
    "egg_0",
    "egg_1",
    "egg_3",
    "egg_4",
    "egg_5",
    "egg_6",
    "egg_7",
    "egg_8",
    "egg_9",
    "egg_10",
    "eggplant_0",
    "eggplant_1",
    "eggplant_2",
    "eggplant_3",
    "eggplant_4",
    "fish_0",
    "fish_1",
    "fish_2",
    "fish_3",
    "fish_4",
    "fish_5",
    "fish_6",
    "fish_7",
    "fish_8",
    "fish_9",
    "fork_0",
    "fork_1",
    "fork_2",
    "fork_3",
    "fork_4",
    "fork_5",
    "fork_6",
    "fork_7",
    "fork_8",
    "fork_9",
    "garlic_0",
    "garlic_1",
    "garlic_2",
    "garlic_4",
    "garlic_5",
    "hot_dog_0",
    "hot_dog_1",
    "hot_dog_2",
    "hot_dog_3",
    "hot_dog_4",
    "hot_dog_5",
    "hot_dog_6",
    "jam_0",
    "jam_1",
    "jam_2",
    "jam_3",
    "jam_4",
    "jam_5",
    "jam_6",
    "jug_0",
    "jug_1",
    "jug_2",
    "jug_3",
    "jug_4",
    "jug_5",
    "jug_6",
    "jug_7",
    "jug_8",
    "ketchup_0",
    "ketchup_1",
    "ketchup_2",
    "ketchup_3",
    "ketchup_4",
    "ketchup_7",
    "ketchup_8",
    "ketchup_9",
    "ketchup_10",
    "ketchup_11",
    "kettle_0",
    "kettle_1",
    "kettle_2",
    "kettle_3",
    "kettle_4",
    "kettle_5",
    "kettle_6",
    "kettle_7",
    "kettle_8",
    "kettle_9",
    "kiwi_0",
    "kiwi_1",
    "kiwi_2",
    "kiwi_4",
    "kiwi_5",
    "kiwi_6",
    "kiwi_7",
    "knife_0",
    "knife_1",
    "knife_2",
    "knife_3",
    "knife_4",
    "knife_5",
    "knife_6",
    "knife_7",
    "knife_8",
    "knife_9",
    "ladle_0",
    "ladle_1",
    "ladle_2",
    "ladle_3",
    "ladle_4",
    "lemon_0",
    "lemon_1",
    "lemon_2",
    "lemon_3",
    "lemon_4",
    "lemon_5",
    "lemon_6",
    "lemon_7",
    "lemon_8",
    "lemon_9",
    "lime_0",
    "lime_1",
    "lime_2",
    "lime_3",
    "mango_1",
    "mango_4",
    "mango_5",
    "mango_7",
    "mango_8",
    "mango_9",
    "mango_10",
    "mango_11",
    "mango_12",
    "mango_13",
    "milk_0",
    "milk_1",
    "milk_3",
    "milk_4",
    "milk_5",
    "milk_7",
    "milk_9",
    "milk_10",
    "mug_0",
    "mug_1",
    "mug_2",
    "mug_3",
    "mug_4",
    "mug_5",
    "mug_7",
    "mug_8",
    "mug_9",
    "mug_10",
    "mushroom_0",
    "mushroom_1",
    "mushroom_2",
    "mushroom_3",
    "mushroom_4",
    "mushroom_5",
    "mushroom_6",
    "mushroom_7",
    "mushroom_8",
    "mushroom_9",
    "onion_0",
    "onion_1",
    "onion_2",
    "onion_3",
    "onion_4",
    "onion_6",
    "onion_7",
    "onion_8",
    "onion_9",
    "onion_10",
    "orange_0",
    "orange_1",
    "orange_2",
    "orange_3",
    "orange_4",
    "orange_5",
    "orange_8",
    "orange_9",
    "orange_10",
    "orange_11",
    "pan_0",
    "pan_1",
    "pan_2",
    "pan_3",
    "pan_4",
    "pan_5",
    "pan_6",
    "pan_7",
    "pan_8",
    "pan_9",
    "peach_0",
    "peach_1",
    "peach_2",
    "peach_3",
    "peach_4",
    "peach_5",
    "peach_6",
    "pear_0",
    "pear_1",
    "pear_2",
    "pear_3",
    "pear_5",
    "pear_6",
    "pear_8",
    "pear_10",
    "pear_11",
    "pear_12",
    "plate_0",
    "plate_1",
    "plate_2",
    "plate_3",
    "plate_4",
    "plate_5",
    "plate_7",
    "plate_8",
    "plate_9",
    "plate_10",
    "potato_0",
    "potato_1",
    "potato_2",
    "potato_3",
    "potato_4",
    "potato_5",
    "potato_6",
    "potato_7",
    "potato_8",
    "potato_9",
    "rolling_pin_0",
    "rolling_pin_1",
    "rolling_pin_2",
    "rolling_pin_3",
    "rolling_pin_4",
    "rolling_pin_5",
    "scissors_0",
    "scissors_1",
    "scissors_3",
    "scissors_4",
    "scissors_6",
    "scissors_8",
    "scissors_9",
    "scissors_10",
    "scissors_11",
    "scissors_12",
    "shaker_0",
    "shaker_1",
    "shaker_2",
    "shaker_3",
    "soap_dispenser_0",
    "soap_dispenser_1",
    "soap_dispenser_2",
    "soap_dispenser_3",
    "soap_dispenser_4",
    "soap_dispenser_6",
    "spatula_0",
    "spatula_1",
    "spatula_2",
    "spatula_3",
    "spatula_4",
    "spatula_5",
    "spatula_6",
    "spatula_7",
    "spatula_8",
    "spatula_9",
    "sponge_0",
    "sponge_1",
    "sponge_2",
    "sponge_3",
    "sponge_4",
    "sponge_5",
    "sponge_6",
    "sponge_8",
    "spoon_0",
    "spoon_1",
    "spoon_2",
    "spoon_3",
    "spoon_4",
    "spoon_5",
    "spoon_6",
    "spoon_7",
    "spoon_8",
    "spoon_9",
    "spray_0",
    "spray_1",
    "spray_2",
    "squash_0",
    "squash_1",
    "squash_3",
    "squash_5",
    "squash_7",
    "squash_8",
    "squash_13",
    "squash_15",
    "squash_16",
    "steak_2",
    "steak_3",
    "steak_4",
    "steak_5",
    "steak_6",
    "steak_8",
    "steak_9",
    "steak_10",
    "steak_11",
    "steak_14",
    "sweet_potato_1",
    "sweet_potato_3",
    "tangerine_0",
    "tangerine_1",
    "tangerine_2",
    "tangerine_3",
    "tangerine_4",
    "tangerine_5",
    "tangerine_6",
    "teapot_0",
    "teapot_3",
    "teapot_4",
    "teapot_5",
    "teapot_6",
    "teapot_7",
    "teapot_10",
    "teapot_11",
    "teapot_13",
    "teapot_14",
    "tomato_0",
    "tomato_1",
    "tomato_2",
    "tomato_3",
    "tomato_5",
    "tomato_6",
    "tomato_7",
    "tomato_8",
    "waffle_0",
    "waffle_1",
    "waffle_3",
    "waffle_4",
    "water_bottle_0",
    "water_bottle_1",
    "water_bottle_2",
    "water_bottle_3",
    "water_bottle_4",
    "water_bottle_5",
    "water_bottle_6",
    "water_bottle_7",
    "water_bottle_8",
    "water_bottle_9",
    "wine_1",
    "wine_2",
    "wine_3",
    "wine_4",
    "wine_5",
    "wine_6",
    "wine_8",
    "wine_9",
    "wine_10",
    "wine_11",
    "yogurt_1",
    "yogurt_2",
    "yogurt_3",
]

all_pick_objects_set = [
    "kettle_1",
    "spoon_14",
    "egg_5",
    "kettle_13",
    "cheese_8",
    "lemon_16",
    "water_bottle_17",
    "banana_23",
    "bagel_6",
    "spatula_6",
    "water_bottle_5",
    "beer_8",
    "condiment_0",
    "bread_1",
    "bagel_2",
    "mango_5",
    "can_11",
    "cutting_board_12",
    "fish_0",
    "cheese_6",
    "spoon_17",
    "jug_7",
    "peach_1",
    "cheese_2",
    "avocado_5",
    "bowl_19",
    "wine_2",
    "ketchup_9",
    "tray_1",
    "mango_4",
    "water_bottle_23",
    "cake_4",
    "cucumber_5",
    "fish_7",
    "apple_8",
    "bottled_water_6",
    "donut_13",
    "mushroom_13",
    "fish_3",
    "plate_15",
    "potato_7",
    "plate_13",
    "banana_21",
    "sponge_6",
    "wine_11",
    "bagged_food_8",
    "bowl_17",
    "cereal_5",
    "kiwi_0",
    "boxed_drink_7",
    "donut_1",
    "cutting_board_4",
    "candle_12",
    "chocolate_6",
    "sponge_4",
    "chips_1",
    "cereal_7",
    "lemon_4",
    "coffee_cup_16",
    "broccoli_0",
    "cereal_11",
    "sponge_2",
    "bell_pepper_0",
    "cupcake_17",
    "bar_soap_3",
    "steak_4",
    "boxed_drink_3",
    "sweet_potato_1",
    "plate_3",
    "can_3",
    "sponge_8",
    "milk_3",
    "kettle_19",
    "scissors_10",
    "boxed_food_1",
    "cake_3",
    "mug_18",
    "plate_10",
    "mushroom_7",
    "water_bottle_6",
    "hot_dog_0",
    "squash_16",
    "peach_0",
    "squash_3",
    "jug_4",
    "mango_16",
    "mushroom_12",
    "bread_6",
    "fork_0",
    "bar_3",
    "kettle_0",
    "mango_9",
    "chips_2",
    "wine_6",
    "coffee_cup_17",
    "coffee_cup_21",
    "potato_1",
    "milk_5",
    "wine_3",
    "wine_1",
    "kettle_23",
    "orange_13",
    "orange_0",
    "cup_3",
    "croissant_9",
    "apple_4",
    "spatula_1",
    "mushroom_0",
    "pear_16",
    "water_bottle_1",
    "carrot_4",
    "cheese_4",
    "kettle_9",
    "beer_16",
    "pan_9",
    "cupcake_18",
    "waffle_1",
    "spoon_6",
    "plate_11",
    "milk_9",
    "steak_2",
    "water_bottle_13",
    "orange_3",
    "fork_5",
    "fork_11",
    "cutting_board_16",
    "spray_2",
    "plate_1",
    "lime_2",
    "bottled_drink_0",
    "bowl_12",
    "kettle_17",
    "kettle_4",
    "peach_3",
    "hot_dog_3",
    "mushroom_3",
    "scissors_8",
    "beer_12",
    "alcohol_13",
    "spatula_12",
    "tray_9",
    "bread_21",
    "hot_dog_1",
    "croissant_1",
    "bread_4",
    "banana_5",
    "boxed_drink_2",
    "fork_6",
    "cupcake_9",
    "mushroom_10",
    "potato_13",
    "candle_1",
    "teapot_4",
    "mug_8",
    "beer_0",
    "pan_21",
    "shaker_0",
    "pan_19",
    "peach_5",
    "apple_9",
    "canned_food_8",
    "canned_food_10",
    "ketchup_2",
    "water_bottle_8",
    "yogurt_1",
    "spoon_8",
    "pear_6",
    "cheese_1",
    "spray_0",
    "knife_17",
    "fork_1",
    "sweet_potato_3",
    "scissors_4",
    "potato_12",
    "broccoli_3",
    "plate_14",
    "banana_16",
    "jug_3",
    "bottled_water_13",
    "bell_pepper_2",
    "avocado_2",
    "pan_23",
    "squash_8",
    "coffee_cup_8",
    "banana_6",
    "bowl_16",
    "bread_18",
    "coffee_cup_12",
    "egg_12",
    "water_bottle_18",
    "bread_13",
    "soap_dispenser_2",
    "cake_0",
    "chocolate_11",
    "lemon_8",
    "knife_15",
    "corn_0",
    "teapot_7",
    "water_bottle_7",
    "canned_food_9",
    "cucumber_1",
    "water_bottle_10",
    "garlic_1",
    "jam_2",
    "knife_2",
    "bowl_1",
    "lime_0",
    "donut_9",
    "lemon_2",
    "beer_9",
    "bowl_18",
    "bowl_9",
    "chocolate_3",
    "tomato_8",
    "spatula_13",
    "pan_18",
    "cake_1",
    "ladle_2",
    "tomato_6",
    "teapot_15",
    "bowl_6",
    "spoon_5",
    "wine_12",
    "squash_15",
    "spoon_18",
    "bread_14",
    "lemon_9",
    "water_bottle_12",
    "pan_8",
    "squash_0",
    "ketchup_0",
    "can_13",
    "waffle_0",
    "jam_4",
    "bar_2",
    "tangerine_3",
    "mushroom_11",
    "boxed_drink_0",
    "carrot_1",
    "water_bottle_19",
    "banana_13",
    "ladle_0",
    "beer_7",
    "chips_8",
    "bottled_water_2",
    "chocolate_12",
    "alcohol_7",
    "egg_13",
    "potato_2",
    "bottled_drink_2",
    "candle_13",
    "kiwi_7",
    "can_0",
    "apple_13",
    "cereal_8",
    "pear_18",
    "tangerine_4",
    "plate_12",
    "knife_1",
    "pear_8",
    "bowl_3",
    "onion_0",
    "kettle_3",
    "scissors_0",
    "jug_1",
    "chocolate_13",
    "banana_18",
    "donut_14",
    "croissant_0",
    "cup_5",
    "onion_9",
    "knife_3",
    "garlic_5",
    "bowl_13",
    "chocolate_8",
    "jam_6",
    "baguette_0",
    "cupcake_6",
    "carrot_3",
    "knife_5",
    "tray_11",
    "kettle_8",
    "croissant_4",
    "knife_19",
    "spatula_4",
    "scissors_6",
    "coffee_cup_3",
    "egg_3",
    "plate_17",
    "can_2",
    "canned_food_6",
    "water_bottle_20",
    "bar_10",
    "jug_8",
    "chocolate_7",
    "pear_12",
    "knife_27",
    "kettle_24",
    "hot_dog_4",
    "peach_2",
    "lime_3",
    "mango_12",
    "mushroom_4",
    "bottled_drink_1",
    "donut_7",
    "cutting_board_13",
    "can_17",
    "bottled_water_12",
    "yogurt_3",
    "chocolate_1",
    "cutting_board_5",
    "eggplant_4",
    "scissors_13",
    "spatula_5",
    "bar_4",
    "bowl_11",
    "ketchup_8",
    "knife_6",
    "plate_0",
    "candle_3",
    "fish_1",
    "lemon_10",
    "waffle_4",
    "water_bottle_21",
    "potato_3",
    "candle_10",
    "spoon_22",
    "spatula_9",
    "fork_4",
    "bagel_9",
    "spatula_0",
    "can_12",
    "apple_7",
    "baguette_6",
    "coffee_cup_13",
    "banana_12",
    "knife_16",
    "cupcake_7",
    "potato_4",
    "cereal_13",
    "bread_7",
    "beer_6",
    "alcohol_0",
    "potato_8",
    "rolling_pin_2",
    "bottled_water_3",
    "bread_2",
    "kiwi_6",
    "bagged_food_0",
    "pan_3",
    "banana_14",
    "milk_0",
    "pan_15",
    "pear_13",
    "cereal_3",
    "chips_7",
    "alcohol_14",
    "garlic_2",
    "soap_dispenser_4",
    "mushroom_6",
    "cereal_10",
    "cake_8",
    "bagged_food_5",
    "beer_5",
    "donut_0",
    "bread_11",
    "spoon_9",
    "cucumber_3",
    "boxed_food_7",
    "spoon_12",
    "potato_16",
    "mango_19",
    "lemon_11",
    "chocolate_10",
    "spatula_11",
    "chocolate_4",
    "banana_1",
    "can_6",
    "croissant_2",
    "donut_4",
    "banana_4",
    "bagel_7",
    "corn_1",
    "mug_11",
    "chips_5",
    "corn_4",
    "potato_5",
    "rolling_pin_0",
    "bowl_5",
    "banana_8",
    "knife_8",
    "mug_1",
    "knife_4",
    "scissors_12",
    "mug_15",
    "teapot_0",
    "canned_food_18",
    "alcohol_9",
    "can_4",
    "croissant_8",
    "pan_6",
    "alcohol_4",
    "pear_0",
    "baguette_5",
    "bottled_water_10",
    "mango_18",
    "tomato_3",
    "broccoli_1",
    "donut_12",
    "orange_10",
    "bar_11",
    "ketchup_11",
    "bottled_water_17",
    "coffee_cup_2",
    "avocado_0",
    "candle_4",
    "bottled_drink_4",
    "knife_11",
    "water_bottle_16",
    "pear_11",
    "pan_17",
    "mug_0",
    "wine_4",
    "boxed_food_6",
    "banana_9",
    "onion_4",
    "tangerine_5",
    "chips_3",
    "chocolate_2",
    "teapot_6",
    "cereal_1",
    "can_16",
    "spoon_10",
    "spatula_3",
    "bread_16",
    "tray_2",
    "potato_6",
    "pan_24",
    "onion_10",
    "candle_9",
    "fish_2",
    "egg_9",
    "chocolate_5",
    "apple_11",
    "ladle_3",
    "kettle_20",
    "cereal_9",
    "bottled_water_4",
    "donut_5",
    "cake_6",
    "beer_2",
    "alcohol_1",
    "kettle_5",
    "canned_food_0",
    "rolling_pin_3",
    "pear_5",
    "lemon_1",
    "fish_9",
    "mug_4",
    "peach_4",
    "teapot_5",
    "boxed_food_4",
    "pan_22",
    "tray_10",
    "knife_7",
    "avocado_8",
    "coffee_cup_6",
    "jug_0",
    "teapot_11",
    "banana_2",
    "cutting_board_2",
    "cupcake_2",
    "mug_7",
    "apple_22",
    "pan_1",
    "cereal_0",
    "steak_6",
    "croissant_5",
    "bread_17",
    "orange_11",
    "water_bottle_15",
    "cutting_board_9",
    "squash_5",
    "potato_18",
    "cereal_6",
    "croissant_3",
    "bottled_water_16",
    "kettle_12",
    "squash_7",
    "lemon_5",
    "fork_8",
    "scissors_3",
    "apple_21",
    "kettle_11",
    "spoon_21",
    "knife_0",
    "banana_19",
    "bread_8",
    "bagel_4",
    "mushroom_14",
    "canned_food_15",
    "fish_10",
    "orange_4",
    "canned_food_11",
    "knife_25",
    "steak_10",
    "chips_11",
    "water_bottle_3",
    "cupcake_16",
    "water_bottle_2",
    "bowl_4",
    "milk_10",
    "knife_13",
    "cheese_3",
    "kettle_27",
    "kettle_22",
    "kiwi_5",
    "spray_1",
    "apple_19",
    "spoon_3",
    "lemon_6",
    "orange_9",
    "pan_11",
    "cupcake_20",
    "orange_5",
    "baguette_4",
    "chips_10",
    "egg_0",
    "cutting_board_15",
    "milk_1",
    "can_9",
    "apple_12",
    "fork_3",
    "sponge_1",
    "tray_4",
    "fish_8",
    "spatula_7",
    "apple_15",
    "lemon_17",
    "boxed_drink_1",
    "wine_10",
    "donut_3",
    "coffee_cup_9",
    "can_8",
    "tomato_5",
    "jug_2",
    "alcohol_11",
    "mango_11",
    "plate_7",
    "onion_6",
    "rolling_pin_4",
    "teapot_13",
    "eggplant_0",
    "chips_4",
    "lemon_15",
    "chips_9",
    "mushroom_2",
    "lemon_13",
    "carrot_6",
    "bagged_food_14",
    "coffee_cup_15",
    "cucumber_0",
    "mushroom_8",
    "potato_17",
    "jam_3",
    "milk_4",
    "cupcake_19",
    "pear_10",
    "canned_food_5",
    "cutting_board_1",
    "candle_6",
    "egg_8",
    "carrot_0",
    "bar_8",
    "knife_24",
    "mushroom_15",
    "boxed_drink_5",
    "apple_1",
    "bowl_8",
    "chocolate_14",
    "water_bottle_9",
    "banana_3",
    "kettle_28",
    "jug_6",
    "chocolate_9",
    "mango_14",
    "banana_7",
    "peach_6",
    "alcohol_10",
    "potato_10",
    "cereal_2",
    "ketchup_3",
    "kettle_14",
    "chips_13",
    "egg_10",
    "plate_2",
    "bagged_food_3",
    "onion_1",
    "mango_1",
    "pear_15",
    "bell_pepper_1",
    "can_18",
    "mug_16",
    "fish_5",
    "bread_3",
    "donut_11",
    "bottled_water_7",
    "pan_2",
    "shaker_3",
    "tray_5",
    "coffee_cup_1",
    "bar_soap_4",
    "shaker_1",
    "banana_11",
    "plate_4",
    "kettle_6",
    "mug_10",
    "pan_7",
    "pan_20",
    "water_bottle_0",
    "garlic_0",
    "cupcake_3",
    "croissant_6",
    "tangerine_6",
    "bottled_drink_5",
    "cheese_0",
    "bar_6",
    "candle_8",
    "candle_15",
    "boxed_drink_4",
    "spatula_8",
    "donut_10",
    "beer_19",
    "cutting_board_7",
    "banana_24",
    "pan_4",
    "bagged_food_9",
    "pan_14",
    "ketchup_10",
    "tray_0",
    "bar_soap_1",
    "candle_7",
    "sponge_5",
    "steak_11",
    "bar_soap_0",
    "onion_8",
    "chocolate_0",
    "rolling_pin_1",
    "cake_9",
    "bowl_10",
    "knife_10",
    "chips_16",
    "egg_4",
    "jam_5",
    "chips_14",
    "steak_9",
    "wine_5",
    "spoon_20",
    "mug_17",
    "boxed_food_11",
    "donut_2",
    "can_15",
    "apple_0",
    "cupcake_4",
    "apple_14",
    "carrot_7",
    "steak_8",
    "mango_13",
    "avocado_4",
    "condiment_3",
    "potato_19",
    "baguette_2",
    "teapot_10",
    "chips_6",
    "water_bottle_22",
    "soap_dispenser_0",
    "carrot_2",
    "boxed_food_10",
    "spoon_11",
    "banana_22",
    "hot_dog_2",
    "boxed_food_8",
    "egg_7",
    "waffle_3",
    "fork_10",
    "knife_22",
    "pear_2",
    "milk_7",
    "bagel_3",
    "tray_3",
    "knife_21",
    "cupcake_14",
    "bottled_water_8",
    "baguette_1",
    "tomato_0",
    "cupcake_11",
    "plate_20",
    "pan_5",
    "chips_15",
    "kiwi_4",
    "bowl_7",
    "mug_14",
    "kettle_21",
    "bagel_1",
    "cup_4",
    "apple_18",
    "apple_5",
    "fork_7",
    "mushroom_1",
    "cheese_5",
    "kiwi_2",
    "soap_dispenser_1",
    "ketchup_1",
    "beer_13",
    "bagel_0",
    "mango_10",
    "spatula_2",
    "knife_26",
    "orange_8",
    "avocado_6",
    "orange_2",
    "croissant_10",
    "knife_23",
    "bottled_water_14",
    "steak_5",
    "mug_5",
    "cereal_12",
    "spoon_16",
    "bar_0",
    "mug_2",
    "bread_10",
    "apple_17",
    "mug_9",
    "wine_13",
    "spoon_7",
    "bread_9",
    "donut_6",
    "hot_dog_6",
    "apple_2",
    "wine_9",
    "orange_12",
    "cupcake_5",
    "fish_4",
    "chips_0",
    "coffee_cup_11",
    "banana_10",
    "sponge_0",
    "fish_6",
    "jam_0",
    "apple_3",
    "pan_0",
    "apple_10",
    "potato_0",
    "bottled_water_1",
    "squash_13",
    "coffee_cup_10",
    "beer_11",
    "potato_11",
    "lime_1",
    "pear_19",
    "spoon_15",
    "candle_5",
    "tray_12",
    "bottled_water_11",
    "pear_17",
    "egg_6",
    "canned_food_3",
    "steak_3",
    "mushroom_5",
    "bowl_14",
    "coffee_cup_14",
    "tangerine_0",
    "bell_pepper_3",
    "ladle_1",
    "soap_dispenser_6",
    "plate_9",
    "scissors_9",
    "fork_9",
    "kettle_16",
    "eggplant_3",
    "fork_2",
    "corn_3",
    "boxed_food_2",
    "mug_3",
    "pan_13",
    "cutting_board_0",
    "avocado_1",
    "water_bottle_4",
    "mango_7",
    "knife_20",
    "lemon_12",
    "kettle_18",
    "cake_5",
    "teapot_14",
    "kettle_2",
    "tomato_2",
    "beer_18",
    "orange_1",
    "plate_19",
    "hot_dog_5",
    "egg_1",
    "banana_15",
    "mango_8",
    "onion_3",
    "steak_15",
    "mango_15",
    "tomato_7",
    "spoon_13",
    "coffee_cup_20",
    "eggplant_2",
    "onion_2",
    "coffee_cup_4",
    "beer_1",
    "spoon_4",
    "squash_1",
    "bagel_5",
    "bowl_15",
    "spatula_10",
    "sponge_3",
    "boxed_food_0",
    "bar_9",
    "kettle_7",
    "jam_1",
    "potato_14",
    "knife_12",
    "pear_3",
    "cutting_board_11",
    "bagged_food_4",
    "spoon_2",
    "plate_5",
    "alcohol_2",
    "beer_15",
    "coffee_cup_7",
    "bowl_0",
    "scissors_11",
    "pear_1",
    "bread_20",
    "canned_food_17",
    "alcohol_12",
    "bread_5",
    "shaker_2",
    "candle_0",
    "garlic_4",
    "tray_7",
    "potato_9",
    "canned_food_13",
    "bowl_2",
    "beer_10",
    "cucumber_2",
    "steak_14",
    "cupcake_15",
    "onion_7",
    "tomato_1",
    "donut_15",
    "boxed_food_9",
    "spoon_23",
    "banana_20",
    "yogurt_2",
    "mug_12",
    "egg_11",
    "lemon_0",
    "kettle_10",
    "donut_8",
    "eggplant_1",
    "mug_13",
    "cereal_4",
    "alcohol_5",
    "knife_9",
    "tangerine_2",
    "bar_7",
    "ladle_4",
    "canned_food_4",
    "beer_3",
    "teapot_3",
    "cake_7",
    "soap_dispenser_3",
    "lemon_7",
    "spoon_0",
    "ketchup_4",
    "candle_2",
    "spoon_19",
    "cup_2",
    "ketchup_7",
    "kiwi_1",
    "apple_6",
    "jug_5",
    "apple_20",
    "bagged_food_7",
    "tangerine_1",
    "fish_11",
    "rolling_pin_5",
    "pan_10",
    "cheese_7",
    "lemon_3",
    "knife_18",
    "candle_14",
    "canned_food_14",
    "condiment_2",
    "lemon_14",
    "plate_8",
    "wine_8",
    "spoon_1",
    "bread_0",
    "canned_food_1",
    "kettle_26",
    "cucumber_4",
    "cupcake_12",
    "canned_food_19",
    "bagged_food_11",
    "scissors_1",
    "mushroom_9"
]


import numpy as np

from egogym.components import Object


class TexturesManager:
    def __init__(self, textures_list, np_random):
        self.textures_list = textures_list
        self.np_random = np_random
        sorted_textures = sorted(textures_list)
        self.np_random.shuffle(sorted_textures)
        self.textures_list = sorted_textures
        self.index = -1

    def sample(self, random=False):
        if random:
            index = self.np_random.integers(len(self.textures_list))
        else:
            self.index = (self.index + 1) % len(self.textures_list)
            index = self.index
        return self.textures_list[index]



import numpy as np

# permutation transformation matrix to go from record3d axis to personal camera axis
P = np.array([[-1, 0, 0, 0], [0, 0, -1, 0], [0, -1, 0, 0], [0, 0, 0, 1]])
# end effector transformation matrix to go from personal camera axis to end effector axis. Corrects the 15 degree offset along the x axis
EFT = np.asarray([[1, 0, 0, 0], [0, 1, 0, 0], [0, 0, 1, 0], [0, 0, 0, 1]])


def apply_permutation_transform(matrix):
    return P @ matrix @ P.T


def invert_permutation_transform(matrix):
    return P.T @ matrix @ P


def apply_end_effector_transform(matrix):
    return EFT @ matrix


def invert_end_effector_transform(matrix):
    return EFT.T @ matrix



from functools import partial

import torch
from torch import nn, einsum
import torch.nn.functional as F
import torch.distributed as distributed
from torch.optim import Optimizer
from torch.cuda.amp import autocast

from einops import rearrange, repeat, reduce, pack, unpack

from typing import Callable


def exists(val):
    return val is not None


def default(val, d):
    return val if exists(val) else d


def noop(*args, **kwargs):
    pass


def identity(t):
    return t


def l2norm(t):
    return F.normalize(t, p=2, dim=-1)


def cdist(x, y):
    x2 = reduce(x**2, "b n d -> b n", "sum")
    y2 = reduce(y**2, "b n d -> b n", "sum")
    xy = einsum("b i d, b j d -> b i j", x, y) * -2
    return (rearrange(x2, "b i -> b i 1") + rearrange(y2, "b j -> b 1 j") + xy).sqrt()


def log(t, eps=1e-20):
    return torch.log(t.clamp(min=eps))


def ema_inplace(old, new, decay):
    is_mps = str(old.device).startswith("mps:")

    if not is_mps:
        old.lerp_(new, 1 - decay)
    else:
        old.mul_(decay).add_(new * (1 - decay))


def pack_one(t, pattern):
    return pack([t], pattern)


def unpack_one(t, ps, pattern):
    return unpack(t, ps, pattern)[0]


def uniform_init(*shape):
    t = torch.empty(shape)
    nn.init.kaiming_uniform_(t)
    return t


def gumbel_noise(t):
    noise = torch.zeros_like(t).uniform_(0, 1)
    return -log(-log(noise))


def gumbel_sample(
    logits,
    temperature=1.0,
    stochastic=False,
    straight_through=False,
    reinmax=False,
    dim=-1,
    training=True,
):
    dtype, size = logits.dtype, logits.shape[dim]

    if training and stochastic and temperature > 0:
        sampling_logits = (logits / temperature) + gumbel_noise(logits)
    else:
        sampling_logits = logits

    ind = sampling_logits.argmax(dim=dim)
    one_hot = F.one_hot(ind, size).type(dtype)

    assert not (
        reinmax and not straight_through
    ), "reinmax can only be turned on if using straight through gumbel softmax"

    if not straight_through or temperature <= 0.0 or not training:
        return ind, one_hot

    # use reinmax for better second-order accuracy - https://arxiv.org/abs/2304.08612
    # algorithm 2

    if reinmax:
        π0 = logits.softmax(dim=dim)
        π1 = (one_hot + (logits / temperature).softmax(dim=dim)) / 2
        π1 = ((log(π1) - logits).detach() + logits).softmax(dim=1)
        π2 = 2 * π1 - 0.5 * π0
        one_hot = π2 - π2.detach() + one_hot
    else:
        π1 = (logits / temperature).softmax(dim=dim)
        one_hot = one_hot + π1 - π1.detach()

    return ind, one_hot


def laplace_smoothing(x, n_categories, eps=1e-5, dim=-1):
    denom = x.sum(dim=dim, keepdim=True)
    return (x + eps) / (denom + n_categories * eps)


def sample_vectors(samples, num):
    num_samples, device = samples.shape[0], samples.device
    if num_samples >= num:
        indices = torch.randperm(num_samples, device=device)[:num]
    else:
        indices = torch.randint(0, num_samples, (num,), device=device)

    return samples[indices]


def batched_sample_vectors(samples, num):
    return torch.stack(
        [sample_vectors(sample, num) for sample in samples.unbind(dim=0)], dim=0
    )


def pad_shape(shape, size, dim=0):
    return [size if i == dim else s for i, s in enumerate(shape)]


def sample_multinomial(total_count, probs):
    device = probs.device
    probs = probs.cpu()

    total_count = probs.new_full((), total_count)
    remainder = probs.new_ones(())
    sample = torch.empty_like(probs, dtype=torch.long)

    for i, p in enumerate(probs):
        s = torch.binomial(total_count, p / remainder)
        sample[i] = s
        total_count -= s
        remainder -= p

    return sample.to(device)


def all_gather_sizes(x, dim):
    size = torch.tensor(x.shape[dim], dtype=torch.long, device=x.device)
    all_sizes = [torch.empty_like(size) for _ in range(distributed.get_world_size())]
    distributed.all_gather(all_sizes, size)
    return torch.stack(all_sizes)


def all_gather_variably_sized(x, sizes, dim=0):
    rank = distributed.get_rank()
    all_x = []

    for i, size in enumerate(sizes):
        t = x if i == rank else x.new_empty(pad_shape(x.shape, size, dim))
        distributed.broadcast(t, src=i, async_op=True)
        all_x.append(t)

    distributed.barrier()
    return all_x


def sample_vectors_distributed(local_samples, num):
    local_samples = rearrange(local_samples, "1 ... -> ...")

    rank = distributed.get_rank()
    all_num_samples = all_gather_sizes(local_samples, dim=0)

    if rank == 0:
        samples_per_rank = sample_multinomial(
            num, all_num_samples / all_num_samples.sum()
        )
    else:
        samples_per_rank = torch.empty_like(all_num_samples)

    distributed.broadcast(samples_per_rank, src=0)
    samples_per_rank = samples_per_rank.tolist()

    local_samples = sample_vectors(local_samples, samples_per_rank[rank])
    all_samples = all_gather_variably_sized(local_samples, samples_per_rank, dim=0)
    out = torch.cat(all_samples, dim=0)

    return rearrange(out, "... -> 1 ...")


def batched_bincount(x, *, minlength):
    batch, dtype, device = x.shape[0], x.dtype, x.device
    target = torch.zeros(batch, minlength, dtype=dtype, device=device)
    values = torch.ones_like(x)
    target.scatter_add_(-1, x, values)
    return target


def kmeans(
    samples,
    num_clusters,
    num_iters=10,
    use_cosine_sim=False,
    sample_fn=batched_sample_vectors,
    all_reduce_fn=noop,
):
    num_codebooks, dim, dtype, device = (
        samples.shape[0],
        samples.shape[-1],
        samples.dtype,
        samples.device,
    )

    means = sample_fn(samples, num_clusters)

    for _ in range(num_iters):
        if use_cosine_sim:
            dists = samples @ rearrange(means, "h n d -> h d n")
        else:
            dists = -torch.cdist(samples, means, p=2)

        buckets = torch.argmax(dists, dim=-1)
        bins = batched_bincount(buckets, minlength=num_clusters)
        all_reduce_fn(bins)

        zero_mask = bins == 0
        bins_min_clamped = bins.masked_fill(zero_mask, 1)

        new_means = buckets.new_zeros(num_codebooks, num_clusters, dim, dtype=dtype)

        new_means.scatter_add_(1, repeat(buckets, "h n -> h n d", d=dim), samples)
        new_means = new_means / rearrange(bins_min_clamped, "... -> ... 1")
        all_reduce_fn(new_means)

        if use_cosine_sim:
            new_means = l2norm(new_means)

        means = torch.where(rearrange(zero_mask, "... -> ... 1"), means, new_means)

    return means, bins


def batched_embedding(indices, embeds):
    batch, dim = indices.shape[1], embeds.shape[-1]
    indices = repeat(indices, "h b n -> h b n d", d=dim)
    embeds = repeat(embeds, "h c d -> h b c d", b=batch)
    return embeds.gather(2, indices)


# regularization losses


def orthogonal_loss_fn(t):
    # eq (2) from https://arxiv.org/abs/2112.00384
    h, n = t.shape[:2]
    normed_codes = l2norm(t)
    cosine_sim = einsum("h i d, h j d -> h i j", normed_codes, normed_codes)
    return (cosine_sim**2).sum() / (h * n**2) - (1 / n)


# distance types


class EuclideanCodebook(nn.Module):
    def __init__(
        self,
        dim,
        codebook_size,
        num_codebooks=1,
        kmeans_init=False,
        kmeans_iters=10,
        sync_kmeans=True,
        decay=0.8,
        eps=1e-5,
        threshold_ema_dead_code=2,
        reset_cluster_size=None,
        use_ddp=False,
        learnable_codebook=False,
        gumbel_sample=gumbel_sample,
        sample_codebook_temp=1.0,
        ema_update=True,
        affine_param=False,
        sync_affine_param=False,
        affine_param_batch_decay=0.99,
        affine_param_codebook_decay=0.9,
    ):
        super().__init__()
        self.transform_input = identity

        self.decay = decay
        self.ema_update = ema_update

        init_fn = uniform_init if not kmeans_init else torch.zeros
        embed = init_fn(num_codebooks, codebook_size, dim)

        self.codebook_size = codebook_size
        self.num_codebooks = num_codebooks

        self.kmeans_iters = kmeans_iters
        self.eps = eps
        self.threshold_ema_dead_code = threshold_ema_dead_code
        self.reset_cluster_size = default(reset_cluster_size, threshold_ema_dead_code)

        assert callable(gumbel_sample)
        self.gumbel_sample = gumbel_sample
        self.sample_codebook_temp = sample_codebook_temp

        assert not (
            use_ddp and num_codebooks > 1 and kmeans_init
        ), "kmeans init is not compatible with multiple codebooks in distributed environment for now"

        self.sample_fn = (
            sample_vectors_distributed
            if use_ddp and sync_kmeans
            else batched_sample_vectors
        )
        self.kmeans_all_reduce_fn = (
            distributed.all_reduce if use_ddp and sync_kmeans else noop
        )
        self.all_reduce_fn = distributed.all_reduce if use_ddp else noop

        self.register_buffer("initted", torch.Tensor([not kmeans_init]))
        self.register_buffer("cluster_size", torch.zeros(num_codebooks, codebook_size))
        self.register_buffer("embed_avg", embed.clone())

        self.learnable_codebook = learnable_codebook
        if learnable_codebook:
            self.embed = nn.Parameter(embed)
        else:
            self.register_buffer("embed", embed)

        # affine related params

        self.affine_param = affine_param
        self.sync_affine_param = sync_affine_param

        if not affine_param:
            return

        self.affine_param_batch_decay = affine_param_batch_decay
        self.affine_param_codebook_decay = affine_param_codebook_decay

        self.register_buffer("batch_mean", None)
        self.register_buffer("batch_variance", None)

        self.register_buffer("codebook_mean_needs_init", torch.Tensor([True]))
        self.register_buffer("codebook_mean", torch.empty(num_codebooks, 1, dim))
        self.register_buffer("codebook_variance_needs_init", torch.Tensor([True]))
        self.register_buffer("codebook_variance", torch.empty(num_codebooks, 1, dim))

    @torch.jit.ignore
    def init_embed_(self, data, mask=None):
        if self.initted:
            return

        if exists(mask):
            c = data.shape[0]
            data = rearrange(data[mask], "(c n) d -> c n d", c=c)

        embed, cluster_size = kmeans(
            data,
            self.codebook_size,
            self.kmeans_iters,
            sample_fn=self.sample_fn,
            all_reduce_fn=self.kmeans_all_reduce_fn,
        )

        embed_sum = embed * rearrange(cluster_size, "... -> ... 1")

        self.embed.data.copy_(embed)
        self.embed_avg.data.copy_(embed_sum)
        self.cluster_size.data.copy_(cluster_size)
        self.initted.data.copy_(torch.Tensor([True]))

    @torch.jit.ignore
    def update_with_decay(self, buffer_name, new_value, decay):
        old_value = getattr(self, buffer_name)

        needs_init = getattr(self, buffer_name + "_needs_init", False)

        if needs_init:
            self.register_buffer(buffer_name + "_needs_init", torch.Tensor([False]))

        if not exists(old_value) or needs_init:
            self.register_buffer(buffer_name, new_value.detach())

            return

        value = old_value * decay + new_value.detach() * (1 - decay)
        self.register_buffer(buffer_name, value)

    @torch.jit.ignore
    def update_affine(self, data, embed, mask=None):
        assert self.affine_param

        var_fn = partial(torch.var, unbiased=False)

        # calculate codebook mean and variance

        embed = rearrange(embed, "h ... d -> h (...) d")

        if self.training:
            self.update_with_decay(
                "codebook_mean",
                reduce(embed, "h n d -> h 1 d", "mean"),
                self.affine_param_codebook_decay,
            )
            self.update_with_decay(
                "codebook_variance",
                reduce(embed, "h n d -> h 1 d", var_fn),
                self.affine_param_codebook_decay,
            )

        # prepare batch data, which depends on whether it has masking

        data = rearrange(data, "h ... d -> h (...) d")

        if exists(mask):
            c = data.shape[0]
            data = rearrange(data[mask], "(c n) d -> c n d", c=c)

        # calculate batch mean and variance

        if not self.sync_affine_param:
            self.update_with_decay(
                "batch_mean",
                reduce(data, "h n d -> h 1 d", "mean"),
                self.affine_param_batch_decay,
            )
            self.update_with_decay(
                "batch_variance",
                reduce(data, "h n d -> h 1 d", var_fn),
                self.affine_param_batch_decay,
            )
            return

        num_vectors, device, dtype = data.shape[-2], data.device, data.dtype

        # number of vectors, for denominator

        num_vectors = torch.tensor([num_vectors], device=device, dtype=dtype)
        distributed.all_reduce(num_vectors)

        # calculate distributed mean

        batch_sum = reduce(data, "h n d -> h 1 d", "sum")
        distributed.all_reduce(batch_sum)
        batch_mean = batch_sum / num_vectors

        self.update_with_decay("batch_mean", batch_mean, self.affine_param_batch_decay)

        # calculate distributed variance

        variance_numer = reduce((data - batch_mean) ** 2, "h n d -> h 1 d", "sum")
        distributed.all_reduce(variance_numer)
        batch_variance = variance_numer / num_vectors

        self.update_with_decay(
            "batch_variance", batch_variance, self.affine_param_batch_decay
        )

    def replace(self, batch_samples, batch_mask):
        for ind, (samples, mask) in enumerate(
            zip(batch_samples.unbind(dim=0), batch_mask.unbind(dim=0))
        ):
            if not torch.any(mask):
                continue

            sampled = self.sample_fn(
                rearrange(samples, "... -> 1 ..."), mask.sum().item()
            )
            sampled = rearrange(sampled, "1 ... -> ...")

            self.embed.data[ind][mask] = sampled

            self.cluster_size.data[ind][mask] = self.reset_cluster_size
            self.embed_avg.data[ind][mask] = sampled * self.reset_cluster_size

    def expire_codes_(self, batch_samples):
        if self.threshold_ema_dead_code == 0:
            return

        expired_codes = self.cluster_size < self.threshold_ema_dead_code

        if not torch.any(expired_codes):
            return

        batch_samples = rearrange(batch_samples, "h ... d -> h (...) d")
        self.replace(batch_samples, batch_mask=expired_codes)

    @autocast(enabled=False)
    def forward(self, x, sample_codebook_temp=None, mask=None, freeze_codebook=False):
        needs_codebook_dim = x.ndim < 4
        sample_codebook_temp = default(sample_codebook_temp, self.sample_codebook_temp)

        x = x.float()

        if needs_codebook_dim:
            x = rearrange(x, "... -> 1 ...")

        dtype = x.dtype
        flatten, ps = pack_one(x, "h * d")

        if exists(mask):
            mask = repeat(
                mask,
                "b n -> c (b h n)",
                c=flatten.shape[0],
                h=flatten.shape[-2] // (mask.shape[0] * mask.shape[1]),
            )

        self.init_embed_(flatten, mask=mask)

        if self.affine_param:
            self.update_affine(flatten, self.embed, mask=mask)

        embed = self.embed if self.learnable_codebook else self.embed.detach()

        if self.affine_param:
            codebook_std = self.codebook_variance.clamp(min=1e-5).sqrt()
            batch_std = self.batch_variance.clamp(min=1e-5).sqrt()
            embed = (embed - self.codebook_mean) * (
                batch_std / codebook_std
            ) + self.batch_mean

        dist = -cdist(flatten, embed)

        embed_ind, embed_onehot = self.gumbel_sample(
            dist, dim=-1, temperature=sample_codebook_temp, training=self.training
        )

        embed_ind = unpack_one(embed_ind, ps, "h *")

        if self.training:
            unpacked_onehot = unpack_one(embed_onehot, ps, "h * c")
            quantize = einsum("h b n c, h c d -> h b n d", unpacked_onehot, embed)
        else:
            quantize = batched_embedding(embed_ind, embed)

        if self.training and self.ema_update and not freeze_codebook:
            if self.affine_param:
                flatten = (flatten - self.batch_mean) * (
                    codebook_std / batch_std
                ) + self.codebook_mean

            if exists(mask):
                embed_onehot[~mask] = 0.0

            cluster_size = embed_onehot.sum(dim=1)

            self.all_reduce_fn(cluster_size)
            ema_inplace(self.cluster_size.data, cluster_size, self.decay)

            embed_sum = einsum("h n d, h n c -> h c d", flatten, embed_onehot)
            self.all_reduce_fn(embed_sum.contiguous())
            ema_inplace(self.embed_avg.data, embed_sum, self.decay)

            cluster_size = laplace_smoothing(
                self.cluster_size, self.codebook_size, self.eps
            ) * self.cluster_size.sum(dim=-1, keepdim=True)

            embed_normalized = self.embed_avg / rearrange(cluster_size, "... -> ... 1")
            self.embed.data.copy_(embed_normalized)
            self.expire_codes_(x)

        if needs_codebook_dim:
            quantize, embed_ind = map(
                lambda t: rearrange(t, "1 ... -> ..."), (quantize, embed_ind)
            )

        dist = unpack_one(dist, ps, "h * d")

        return quantize, embed_ind, dist


class CosineSimCodebook(nn.Module):
    def __init__(
        self,
        dim,
        codebook_size,
        num_codebooks=1,
        kmeans_init=False,
        kmeans_iters=10,
        sync_kmeans=True,
        decay=0.8,
        eps=1e-5,
        threshold_ema_dead_code=2,
        reset_cluster_size=None,
        use_ddp=False,
        learnable_codebook=False,
        gumbel_sample=gumbel_sample,
        sample_codebook_temp=1.0,
        ema_update=True,
    ):
        super().__init__()
        self.transform_input = l2norm

        self.ema_update = ema_update
        self.decay = decay

        if not kmeans_init:
            embed = l2norm(uniform_init(num_codebooks, codebook_size, dim))
        else:
            embed = torch.zeros(num_codebooks, codebook_size, dim)

        self.codebook_size = codebook_size
        self.num_codebooks = num_codebooks

        self.kmeans_iters = kmeans_iters
        self.eps = eps
        self.threshold_ema_dead_code = threshold_ema_dead_code
        self.reset_cluster_size = default(reset_cluster_size, threshold_ema_dead_code)

        assert callable(gumbel_sample)
        self.gumbel_sample = gumbel_sample
        self.sample_codebook_temp = sample_codebook_temp

        self.sample_fn = (
            sample_vectors_distributed
            if use_ddp and sync_kmeans
            else batched_sample_vectors
        )
        self.kmeans_all_reduce_fn = (
            distributed.all_reduce if use_ddp and sync_kmeans else noop
        )
        self.all_reduce_fn = distributed.all_reduce if use_ddp else noop

        self.register_buffer("initted", torch.Tensor([not kmeans_init]))
        self.register_buffer("cluster_size", torch.zeros(num_codebooks, codebook_size))
        self.register_buffer("embed_avg", embed.clone())

        self.learnable_codebook = learnable_codebook
        if learnable_codebook:
            self.embed = nn.Parameter(embed)
        else:
            self.register_buffer("embed", embed)

    @torch.jit.ignore
    def init_embed_(self, data, mask=None):
        if self.initted:
            return

        if exists(mask):
            c = data.shape[0]
            data = rearrange(data[mask], "(c n) d -> c n d", c=c)

        embed, cluster_size = kmeans(
            data,
            self.codebook_size,
            self.kmeans_iters,
            use_cosine_sim=True,
            sample_fn=self.sample_fn,
            all_reduce_fn=self.kmeans_all_reduce_fn,
        )

        embed_sum = embed * rearrange(cluster_size, "... -> ... 1")

        self.embed.data.copy_(embed)
        self.embed_avg.data.copy_(embed_sum)
        self.cluster_size.data.copy_(cluster_size)
        self.initted.data.copy_(torch.Tensor([True]))

    def replace(self, batch_samples, batch_mask):
        batch_samples = l2norm(batch_samples)

        for ind, (samples, mask) in enumerate(
            zip(batch_samples.unbind(dim=0), batch_mask.unbind(dim=0))
        ):
            if not torch.any(mask):
                continue

            sampled = self.sample_fn(
                rearrange(samples, "... -> 1 ..."), mask.sum().item()
            )
            sampled = rearrange(sampled, "1 ... -> ...")

            self.embed.data[ind][mask] = sampled
            self.embed_avg.data[ind][mask] = sampled * self.reset_cluster_size
            self.cluster_size.data[ind][mask] = self.reset_cluster_size

    def expire_codes_(self, batch_samples):
        if self.threshold_ema_dead_code == 0:
            return

        expired_codes = self.cluster_size < self.threshold_ema_dead_code

        if not torch.any(expired_codes):
            return

        batch_samples = rearrange(batch_samples, "h ... d -> h (...) d")
        self.replace(batch_samples, batch_mask=expired_codes)

    @autocast(enabled=False)
    def forward(self, x, sample_codebook_temp=None, mask=None, freeze_codebook=False):
        needs_codebook_dim = x.ndim < 4
        sample_codebook_temp = default(sample_codebook_temp, self.sample_codebook_temp)

        x = x.float()

        if needs_codebook_dim:
            x = rearrange(x, "... -> 1 ...")

        dtype = x.dtype

        flatten, ps = pack_one(x, "h * d")

        if exists(mask):
            mask = repeat(
                mask,
                "b n -> c (b h n)",
                c=flatten.shape[0],
                h=flatten.shape[-2] // (mask.shape[0] * mask.shape[1]),
            )

        self.init_embed_(flatten, mask=mask)

        embed = self.embed if self.learnable_codebook else self.embed.detach()

        dist = einsum("h n d, h c d -> h n c", flatten, embed)

        embed_ind, embed_onehot = self.gumbel_sample(
            dist, dim=-1, temperature=sample_codebook_temp, training=self.training
        )
        embed_ind = unpack_one(embed_ind, ps, "h *")

        if self.training:
            unpacked_onehot = unpack_one(embed_onehot, ps, "h * c")
            quantize = einsum("h b n c, h c d -> h b n d", unpacked_onehot, embed)
        else:
            quantize = batched_embedding(embed_ind, embed)

        if self.training and self.ema_update and not freeze_codebook:
            if exists(mask):
                embed_onehot[~mask] = 0.0

            bins = embed_onehot.sum(dim=1)
            self.all_reduce_fn(bins)

            ema_inplace(self.cluster_size.data, bins, self.decay)

            embed_sum = einsum("h n d, h n c -> h c d", flatten, embed_onehot)
            self.all_reduce_fn(embed_sum.contiguous())
            ema_inplace(self.embed_avg.data, embed_sum, self.decay)

            cluster_size = laplace_smoothing(
                self.cluster_size, self.codebook_size, self.eps
            ) * self.cluster_size.sum(dim=-1, keepdim=True)

            embed_normalized = self.embed_avg / rearrange(cluster_size, "... -> ... 1")
            embed_normalized = l2norm(embed_normalized)

            self.embed.data.copy_(l2norm(embed_normalized))
            self.expire_codes_(x)

        if needs_codebook_dim:
            quantize, embed_ind = map(
                lambda t: rearrange(t, "1 ... -> ..."), (quantize, embed_ind)
            )

        dist = unpack_one(dist, ps, "h * d")
        return quantize, embed_ind, dist


# main class


class VectorQuantize(nn.Module):
    def __init__(
        self,
        dim,
        codebook_size,
        codebook_dim=None,
        heads=1,
        separate_codebook_per_head=False,
        decay=0.8,
        eps=1e-5,
        freeze_codebook=False,
        kmeans_init=True,
        kmeans_iters=10,
        sync_kmeans=True,
        use_cosine_sim=False,
        threshold_ema_dead_code=2,
        channel_last=True,
        accept_image_fmap=False,
        commitment_weight=1.0,
        commitment_use_cross_entropy_loss=False,
        orthogonal_reg_weight=0.0,
        orthogonal_reg_active_codes_only=False,
        orthogonal_reg_max_codes=None,
        stochastic_sample_codes=False,
        sample_codebook_temp=1.0,
        straight_through=False,
        reinmax=False,  # using reinmax for improved straight-through, assuming straight through helps at all
        sync_codebook=None,
        sync_affine_param=False,
        ema_update=True,
        learnable_codebook=False,
        in_place_codebook_optimizer: Callable[
            ..., Optimizer
        ] = None,  # Optimizer used to update the codebook embedding if using learnable_codebook
        affine_param=False,
        affine_param_batch_decay=0.99,
        affine_param_codebook_decay=0.9,
        sync_update_v=0.0,  # the v that controls optimistic vs pessimistic update for synchronous update rule (21) https://minyoungg.github.io/vqtorch/assets/draft_050523.pdf
        eval=False,
    ):
        super().__init__()
        if eval:
            kmeans_init = False
            ema_update = False
        self.dim = dim
        self.heads = heads
        self.separate_codebook_per_head = separate_codebook_per_head

        codebook_dim = default(codebook_dim, dim)
        codebook_input_dim = codebook_dim * heads

        requires_projection = codebook_input_dim != dim
        self.project_in = (
            nn.Linear(dim, codebook_input_dim) if requires_projection else nn.Identity()
        )
        self.project_out = (
            nn.Linear(codebook_input_dim, dim) if requires_projection else nn.Identity()
        )

        self.eps = eps
        self.commitment_weight = commitment_weight
        self.commitment_use_cross_entropy_loss = commitment_use_cross_entropy_loss  # whether to use cross entropy loss to codebook as commitment loss

        self.learnable_codebook = learnable_codebook

        has_codebook_orthogonal_loss = orthogonal_reg_weight > 0
        self.has_codebook_orthogonal_loss = has_codebook_orthogonal_loss
        self.orthogonal_reg_weight = orthogonal_reg_weight
        self.orthogonal_reg_active_codes_only = orthogonal_reg_active_codes_only
        self.orthogonal_reg_max_codes = orthogonal_reg_max_codes

        assert not (
            ema_update and learnable_codebook
        ), "learnable codebook not compatible with EMA update"

        assert 0 <= sync_update_v <= 1.0
        assert not (
            sync_update_v > 0.0 and not learnable_codebook
        ), "learnable codebook must be turned on"

        self.sync_update_v = sync_update_v

        codebook_class = EuclideanCodebook if not use_cosine_sim else CosineSimCodebook

        gumbel_sample_fn = partial(
            gumbel_sample,
            stochastic=stochastic_sample_codes,
            reinmax=reinmax,
            straight_through=straight_through,
        )

        if not exists(sync_codebook):
            sync_codebook = (
                distributed.is_initialized() and distributed.get_world_size() > 1
            )

        codebook_kwargs = dict(
            dim=codebook_dim,
            num_codebooks=heads if separate_codebook_per_head else 1,
            codebook_size=codebook_size,
            kmeans_init=kmeans_init,
            kmeans_iters=kmeans_iters,
            sync_kmeans=sync_kmeans,
            decay=decay,
            eps=eps,
            threshold_ema_dead_code=threshold_ema_dead_code,
            use_ddp=sync_codebook,
            learnable_codebook=has_codebook_orthogonal_loss or learnable_codebook,
            sample_codebook_temp=sample_codebook_temp,
            gumbel_sample=gumbel_sample_fn,
            ema_update=ema_update,
        )

        if affine_param:
            assert (
                not use_cosine_sim
            ), "affine param is only compatible with euclidean codebook"
            codebook_kwargs = dict(
                **codebook_kwargs,
                affine_param=True,
                sync_affine_param=sync_affine_param,
                affine_param_batch_decay=affine_param_batch_decay,
                affine_param_codebook_decay=affine_param_codebook_decay,
            )

        self._codebook = codebook_class(**codebook_kwargs)
        if eval:
            self._codebook.eval()

        self.in_place_codebook_optimizer = (
            in_place_codebook_optimizer(self._codebook.parameters())
            if exists(in_place_codebook_optimizer)
            else None
        )

        self.codebook_size = codebook_size

        self.accept_image_fmap = accept_image_fmap
        self.channel_last = channel_last

    @property
    def codebook(self):
        codebook = self._codebook.embed

        if self.separate_codebook_per_head:
            return codebook

        return rearrange(codebook, "1 ... -> ...")

    @codebook.setter
    def codebook(self, codes):
        if not self.separate_codebook_per_head:
            codes = rearrange(codes, "... -> 1 ...")

        self._codebook.embed.copy_(codes)

    def get_codes_from_indices(self, indices):
        codebook = self.codebook
        is_multiheaded = codebook.ndim > 2

        if not is_multiheaded:
            codes = codebook[indices]
            return rearrange(codes, "... h d -> ... (h d)")

        indices, ps = pack_one(indices, "b * h")
        indices = rearrange(indices, "b n h -> b h n")

        indices = repeat(indices, "b h n -> b h n d", d=codebook.shape[-1])
        codebook = repeat(codebook, "h n d -> b h n d", b=indices.shape[0])

        codes = codebook.gather(2, indices)
        codes = rearrange(codes, "b h n d -> b n (h d)")
        codes = unpack_one(codes, ps, "b * d")
        return codes

    def forward(
        self,
        x,
        indices=None,
        mask=None,
        sample_codebook_temp=None,
        freeze_codebook=False,
    ):
        orig_input = x

        only_one = x.ndim == 2

        if only_one:
            assert not exists(mask)
            x = rearrange(x, "b d -> b 1 d")

        shape, device, heads, is_multiheaded, codebook_size, return_loss = (
            x.shape,
            x.device,
            self.heads,
            self.heads > 1,
            self.codebook_size,
            exists(indices),
        )

        need_transpose = not self.channel_last and not self.accept_image_fmap
        should_inplace_optimize = exists(self.in_place_codebook_optimizer)

        # rearrange inputs

        if self.accept_image_fmap:
            height, width = x.shape[-2:]
            x = rearrange(x, "b c h w -> b (h w) c")

        if need_transpose:
            x = rearrange(x, "b d n -> b n d")

        # project input

        x = self.project_in(x)

        # handle multi-headed separate codebooks

        if is_multiheaded:
            ein_rhs_eq = "h b n d" if self.separate_codebook_per_head else "1 (b h) n d"
            x = rearrange(x, f"b n (h d) -> {ein_rhs_eq}", h=heads)

        # l2norm for cosine sim, otherwise identity

        x = self._codebook.transform_input(x)

        # codebook forward kwargs

        codebook_forward_kwargs = dict(
            sample_codebook_temp=sample_codebook_temp,
            mask=mask,
            freeze_codebook=freeze_codebook,
        )

        # quantize

        quantize, embed_ind, distances = self._codebook(x, **codebook_forward_kwargs)

        # one step in-place update

        if should_inplace_optimize and self.training and not freeze_codebook:
            if exists(mask):
                loss = F.mse_loss(quantize, x.detach(), reduction="none")

                loss_mask = mask
                if is_multiheaded:
                    loss_mask = repeat(
                        mask,
                        "b n -> c (b h) n",
                        c=loss.shape[0],
                        h=loss.shape[1] // mask.shape[0],
                    )

                loss = loss[loss_mask].mean()

            else:
                loss = F.mse_loss(quantize, x.detach())

            loss.backward()
            self.in_place_codebook_optimizer.step()
            self.in_place_codebook_optimizer.zero_grad()

            # quantize again

            quantize, embed_ind, distances = self._codebook(
                x, **codebook_forward_kwargs
            )

        if self.training:
            # determine code to use for commitment loss
            maybe_detach = (
                torch.detach
                if not self.learnable_codebook or freeze_codebook
                else identity
            )

            commit_quantize = maybe_detach(quantize)

            # straight through

            quantize = x + (quantize - x).detach()

            if self.sync_update_v > 0.0:
                # (21) in https://minyoungg.github.io/vqtorch/assets/draft_050523.pdf
                quantize = quantize + self.sync_update_v * (
                    quantize - quantize.detach()
                )

        # function for calculating cross entropy loss to distance matrix
        # used for (1) naturalspeech2 training residual vq latents to be close to the correct codes and (2) cross-entropy based commitment loss

        def calculate_ce_loss(codes):
            if not is_multiheaded:
                dist_einops_eq = "1 b n l -> b l n"
            elif self.separate_codebook_per_head:
                dist_einops_eq = "c b n l -> b l n c"
            else:
                dist_einops_eq = "1 (b h) n l -> b l n h"

            ce_loss = F.cross_entropy(
                rearrange(distances, dist_einops_eq, b=shape[0]), codes, ignore_index=-1
            )

            return ce_loss

        # if returning cross entropy loss on codes that were passed in

        if return_loss:
            return quantize, calculate_ce_loss(indices)

        # transform embedding indices

        if is_multiheaded:
            if self.separate_codebook_per_head:
                embed_ind = rearrange(embed_ind, "h b n -> b n h", h=heads)
            else:
                embed_ind = rearrange(embed_ind, "1 (b h) n -> b n h", h=heads)

        if self.accept_image_fmap:
            embed_ind = rearrange(
                embed_ind, "b (h w) ... -> b h w ...", h=height, w=width
            )

        if only_one:
            embed_ind = rearrange(embed_ind, "b 1 -> b")

        # aggregate loss

        loss = torch.tensor([0.0], device=device, requires_grad=self.training)

        if self.training:
            if self.commitment_weight > 0:
                if self.commitment_use_cross_entropy_loss:
                    if exists(mask):
                        ce_loss_mask = mask
                        if is_multiheaded:
                            ce_loss_mask = repeat(ce_loss_mask, "b n -> b n h", h=heads)

                        embed_ind.masked_fill_(~ce_loss_mask, -1)

                    commit_loss = calculate_ce_loss(embed_ind)
                else:
                    if exists(mask):
                        # with variable lengthed sequences
                        commit_loss = F.mse_loss(commit_quantize, x, reduction="none")

                        loss_mask = mask
                        if is_multiheaded:
                            loss_mask = repeat(
                                loss_mask,
                                "b n -> c (b h) n",
                                c=commit_loss.shape[0],
                                h=commit_loss.shape[1] // mask.shape[0],
                            )

                        commit_loss = commit_loss[loss_mask].mean()
                    else:
                        commit_loss = F.mse_loss(commit_quantize, x)

                loss = loss + commit_loss * self.commitment_weight

            if self.has_codebook_orthogonal_loss:
                codebook = self._codebook.embed

                # only calculate orthogonal loss for the activated codes for this batch

                if self.orthogonal_reg_active_codes_only:
                    assert not (
                        is_multiheaded and self.separate_codebook_per_head
                    ), "orthogonal regularization for only active codes not compatible with multi-headed with separate codebooks yet"
                    unique_code_ids = torch.unique(embed_ind)
                    codebook = codebook[:, unique_code_ids]

                num_codes = codebook.shape[-2]

                if (
                    exists(self.orthogonal_reg_max_codes)
                    and num_codes > self.orthogonal_reg_max_codes
                ):
                    rand_ids = torch.randperm(num_codes, device=device)[
                        : self.orthogonal_reg_max_codes
                    ]
                    codebook = codebook[:, rand_ids]

                orthogonal_reg_loss = orthogonal_loss_fn(codebook)
                loss = loss + orthogonal_reg_loss * self.orthogonal_reg_weight

        # handle multi-headed quantized embeddings

        if is_multiheaded:
            if self.separate_codebook_per_head:
                quantize = rearrange(quantize, "h b n d -> b n (h d)", h=heads)
            else:
                quantize = rearrange(quantize, "1 (b h) n d -> b n (h d)", h=heads)

        # project out

        quantize = self.project_out(quantize)

        # rearrange quantized embeddings

        if need_transpose:
            quantize = rearrange(quantize, "b n d -> b d n")

        if self.accept_image_fmap:
            quantize = rearrange(quantize, "b (h w) c -> b c h w", h=height, w=width)

        if only_one:
            quantize = rearrange(quantize, "b 1 d -> b d")

        # if masking, only return quantized for where mask has True

        if exists(mask):
            quantize = torch.where(
                rearrange(mask, "... -> ... 1"), quantize, orig_input
            )

        return quantize, embed_ind, loss



import os
import numpy as np
from scipy.spatial.transform import Rotation as R
from quaternion import (
    as_rotation_matrix,
    quaternion,
)
import torch
P = np.array([[-1, 0, 0, 0], [0, 0, -1, 0], [0, -1, 0, 0], [0, 0, 0, 1]])

def apply_permutation_transform(matrix):
    return P @ matrix @ P.T

class SimpleReplay:
    def __init__(self, poses_file_path, timeskip=8):
        self.poses_file_path = poses_file_path
        self.timeskip = timeskip
        self.idx = 0
        self.transforms = None

        self.process_poses()

    def to(self, device):
        pass

    def eval(self):
        pass

    def get_poses(self):
        with open(self.poses_file_path, "r") as f:
            lines = f.readlines()
        poses, timestamps = [], []
        for line in lines:
            line_list = eval(line)
            ts = int(line_list[0].split("<")[1].split(">")[0])
            pose = np.array(line_list[1:])
            timestamps.append(ts)
            poses.append(pose)
        timestamps = np.array(timestamps)
        poses = np.array(poses)

        return poses, timestamps

    def process_poses(self):
        quaternions = []
        translations = []
        init_pose = None

        poses, timestamps = self.get_poses()
        for pose in poses:
            qx, qy, qz, qw, tx, ty, tz = pose
            ext_matrix = np.eye(4)
            ext_matrix[:3, :3] = as_rotation_matrix(quaternion(qw, qx, qy, qz))
            ext_matrix[:3, 3] = tx, ty, tz

            if init_pose is None:
                init_pose = np.copy(ext_matrix)
            relative_pose = np.linalg.inv(init_pose) @ ext_matrix

            relative_pose = apply_permutation_transform(relative_pose)
            translations.append(relative_pose[:3, -1])
            quaternions.append(
                R.from_matrix(relative_pose[:3, :3]).as_quat()
            )
        quats = np.array(quaternions)
        translations = np.array(translations)
        transforms = np.concatenate([translations, quats], axis=1)

        self.transforms = transforms
    
    def get_action(self, idx):
        prior_translations, prior_rotations = self.transforms[idx, :3], self.transforms[idx, 3:]
        next_translations, next_rotations = self.transforms[idx + self.timeskip, :3], self.transforms[idx + self.timeskip, 3:]
        # Now, create the matrices.
        prior_rot_matrices, next_rot_matrices = (
            R.from_quat(prior_rotations).as_matrix(),
            R.from_quat(next_rotations).as_matrix(),
        )
        # Now, compute the relative matrices.
        prior_matrices = np.eye(4)
        prior_matrices[:3, :3] = prior_rot_matrices
        prior_matrices[:3, 3] = prior_translations

        next_matrices = np.eye(4)
        next_matrices[:3, :3] = next_rot_matrices
        next_matrices[:3, 3] = next_translations

        relative_transforms = np.matmul(np.linalg.inv(prior_matrices), next_matrices)
        relative_translations = relative_transforms[:3, 3]
        relative_rotations = R.from_matrix(relative_transforms[:3, :3]).as_rotvec()

        gripper = 1.0

        return np.concatenate([relative_translations, relative_rotations, [gripper]], dtype=np.float32)
    
    def step(self, img, step_no):
        start_idx = self.timeskip * step_no
        if start_idx + self.timeskip >= len(self.transforms):
            print("INDEX OUT OF BOUNDS")
            action_tensor = torch.zeros(7)
            action_tensor[-1] = 1
            return action_tensor, {}
        
        action_tensor = self.get_action(start_idx)
        return torch.tensor(action_tensor), {}
    
    def reset(self):
        pass


from egogym.embodiments.grippers.floating_gripper import FloatingGripper

class RUMGripper(FloatingGripper):
    def __init__(self, control_steps=200, grasping_steps=2000, action_space="delta", np_random=None):
        super().__init__(control_steps, grasping_steps, action_space, np_random=np_random)
        self.name = "rum"


from math import ceil
from functools import partial
from itertools import zip_longest
from random import randrange

import torch
from torch import nn
import torch.nn.functional as F
from baselines.rum.models.bet.vqvae.vector_quantize_pytorch import VectorQuantize

from einops import rearrange, repeat, pack, unpack

# helper functions


def exists(val):
    return val is not None


def default(val, d):
    return val if exists(val) else d


def round_up_multiple(num, mult):
    return ceil(num / mult) * mult


# main class


class ResidualVQ(nn.Module):
    """Follows Algorithm 1. in https://arxiv.org/pdf/2107.03312.pdf"""

    def __init__(
        self,
        *,
        dim,
        num_quantizers,
        codebook_dim=None,
        shared_codebook=False,
        heads=1,
        quantize_dropout=False,
        quantize_dropout_cutoff_index=0,
        quantize_dropout_multiple_of=1,
        accept_image_fmap=False,
        eval=False,
        **kwargs,
    ):
        super().__init__()
        assert heads == 1, "residual vq is not compatible with multi-headed codes"
        codebook_dim = default(codebook_dim, dim)
        codebook_input_dim = codebook_dim * heads

        requires_projection = codebook_input_dim != dim
        self.project_in = (
            nn.Linear(dim, codebook_input_dim) if requires_projection else nn.Identity()
        )
        self.project_out = (
            nn.Linear(codebook_input_dim, dim) if requires_projection else nn.Identity()
        )

        self.num_quantizers = num_quantizers

        self.accept_image_fmap = accept_image_fmap
        self.layers = nn.ModuleList(
            [
                VectorQuantize(
                    dim=codebook_dim,
                    codebook_dim=codebook_dim,
                    accept_image_fmap=accept_image_fmap,
                    eval=eval,
                    **kwargs,
                )
                for _ in range(num_quantizers)
            ]
        )
        if eval:
            self.layers.eval()

        self.quantize_dropout = quantize_dropout and num_quantizers > 1

        assert quantize_dropout_cutoff_index >= 0

        self.quantize_dropout_cutoff_index = quantize_dropout_cutoff_index
        self.quantize_dropout_multiple_of = quantize_dropout_multiple_of  # encodec paper proposes structured dropout, believe this was set to 4

        if not shared_codebook:
            return

        first_vq, *rest_vq = self.layers
        codebook = first_vq._codebook

        for vq in rest_vq:
            vq._codebook = codebook

    @property
    def codebooks(self):
        codebooks = [layer._codebook.embed for layer in self.layers]
        codebooks = torch.stack(codebooks, dim=0)
        codebooks = rearrange(codebooks, "q 1 c d -> q c d")
        return codebooks

    def get_codes_from_indices(self, indices):
        batch, quantize_dim = indices.shape[0], indices.shape[-1]

        # may also receive indices in the shape of 'b h w q' (accept_image_fmap)

        indices, ps = pack([indices], "b * q")

        # because of quantize dropout, one can pass in indices that are coarse
        # and the network should be able to reconstruct

        if quantize_dim < self.num_quantizers:
            assert self.quantize_dropout > 0.0, (
                "quantize dropout must be greater than 0 if you wish to reconstruct from a signal with less fine quantizations"
            )
            indices = F.pad(indices, (0, self.num_quantizers - quantize_dim), value=-1)

        # get ready for gathering

        codebooks = repeat(self.codebooks, "q c d -> q b c d", b=batch)
        gather_indices = repeat(indices, "b n q -> q b n d", d=codebooks.shape[-1])

        # take care of quantizer dropout

        mask = gather_indices == -1.0
        gather_indices = gather_indices.masked_fill(
            mask, 0
        )  # have it fetch a dummy code to be masked out later

        all_codes = codebooks.gather(2, gather_indices)  # gather all codes

        # mask out any codes that were dropout-ed

        all_codes = all_codes.masked_fill(mask, 0.0)

        # if (accept_image_fmap = True) then return shape (quantize, batch, height, width, dimension)

        (all_codes,) = unpack(all_codes, ps, "q b * d")

        return all_codes

    def draw_logits_forward(self, encoding_logits):
        # encoding_indices : dim1 = batch_size  dim2 = 4 (number of groups) dim3 = vq dict size (header)
        encoding_logits = encoding_logits.to(self.device)
        bs = encoding_logits.shape[0]
        quantized = torch.zeros((bs, self.codebooks.shape[-1])).to(self.device)
        for q in range(encoding_logits.shape[1]):
            quantized += torch.matmul(encoding_logits[:, q], self.codebooks[q]).to(
                self.device
            )
        return quantized

    def forward(
        self, x, indices=None, return_all_codes=False, sample_codebook_temp=None
    ):
        num_quant, quant_dropout_multiple_of, return_loss, device = (
            self.num_quantizers,
            self.quantize_dropout_multiple_of,
            exists(indices),
            x.device,
        )

        x = self.project_in(x)

        assert not (self.accept_image_fmap and exists(indices))

        quantized_out = 0.0
        residual = x

        all_losses = []
        all_indices = []

        if return_loss:
            assert not torch.any(indices == -1), (
                "some of the residual vq indices were dropped out. please use indices derived when the module is in eval mode to derive cross entropy loss"
            )
            ce_losses = []

        should_quantize_dropout = (
            self.training and self.quantize_dropout and not return_loss
        )

        # sample a layer index at which to dropout further residual quantization
        # also prepare null indices and loss

        if should_quantize_dropout:
            rand_quantize_dropout_index = randrange(
                self.quantize_dropout_cutoff_index, num_quant
            )

            if quant_dropout_multiple_of != 1:
                rand_quantize_dropout_index = (
                    round_up_multiple(
                        rand_quantize_dropout_index + 1, quant_dropout_multiple_of
                    )
                    - 1
                )

            null_indices_shape = (
                (x.shape[0], *x.shape[-2:])
                if self.accept_image_fmap
                else tuple(x.shape[:2])
            )
            null_indices = torch.full(
                null_indices_shape, -1.0, device=device, dtype=torch.long
            )
            null_loss = torch.full((1,), 0.0, device=device, dtype=x.dtype)

        # go through the layers

        for quantizer_index, layer in enumerate(self.layers):
            if (
                should_quantize_dropout
                and quantizer_index > rand_quantize_dropout_index
            ):
                all_indices.append(null_indices)
                all_losses.append(null_loss)
                continue

            layer_indices = None
            if return_loss:
                layer_indices = indices[..., quantizer_index]

            quantized, *rest = layer(
                residual,
                indices=layer_indices,
                sample_codebook_temp=sample_codebook_temp,
            )

            residual = residual - quantized.detach()
            quantized_out = quantized_out + quantized

            if return_loss:
                ce_loss = rest[0]
                ce_losses.append(ce_loss)
                continue

            embed_indices, loss = rest

            all_indices.append(embed_indices)
            all_losses.append(loss)

        # project out, if needed

        quantized_out = self.project_out(quantized_out)

        # whether to early return the cross entropy loss

        if return_loss:
            return quantized_out, sum(ce_losses)

        # stack all losses and indices

        all_losses, all_indices = map(
            partial(torch.stack, dim=-1), (all_losses, all_indices)
        )

        ret = (quantized_out, all_indices, all_losses)

        if return_all_codes:
            # whether to return all codes from all codebooks across layers
            all_codes = self.get_codes_from_indices(all_indices)

            # will return all codes in shape (quantizer, batch, sequence length, codebook dimension)
            ret = (*ret, all_codes)

        return ret


# grouped residual vq


class GroupedResidualVQ(nn.Module):
    def __init__(self, *, dim, groups=1, accept_image_fmap=False, **kwargs):
        super().__init__()
        self.dim = dim
        self.groups = groups
        assert (dim % groups) == 0
        dim_per_group = dim // groups

        self.accept_image_fmap = accept_image_fmap

        self.rvqs = nn.ModuleList([])

        for _ in range(groups):
            self.rvqs.append(
                ResidualVQ(
                    dim=dim_per_group, accept_image_fmap=accept_image_fmap, **kwargs
                )
            )

    @property
    def codebooks(self):
        return torch.stack(tuple(rvq.codebooks for rvq in self.rvqs))

    def get_codes_from_indices(self, indices):
        codes = tuple(
            rvq.get_codes_from_indices(chunk_indices)
            for rvq, chunk_indices in zip(self.rvqs, indices)
        )
        return torch.stack(codes)

    def forward(
        self, x, indices=None, return_all_codes=False, sample_codebook_temp=None
    ):
        shape = x.shape
        split_dim = 1 if self.accept_image_fmap else -1
        assert shape[split_dim] == self.dim

        # split the feature dimension into groups

        x = x.chunk(self.groups, dim=split_dim)

        indices = default(indices, tuple())
        return_ce_loss = len(indices) > 0
        assert len(indices) == 0 or len(indices) == self.groups

        forward_kwargs = dict(
            return_all_codes=return_all_codes, sample_codebook_temp=sample_codebook_temp
        )

        # invoke residual vq on each group

        out = tuple(
            rvq(chunk, indices=chunk_indices, **forward_kwargs)
            for rvq, chunk, chunk_indices in zip_longest(self.rvqs, x, indices)
        )
        out = tuple(zip(*out))

        # if returning cross entropy loss to rvq codebooks

        if return_ce_loss:
            quantized, ce_losses = out
            return torch.cat(quantized, dim=split_dim), sum(ce_losses)

        # otherwise, get all the zipped outputs and combine them

        quantized, all_indices, commit_losses, *maybe_all_codes = out

        quantized = torch.cat(quantized, dim=split_dim)
        all_indices = torch.stack(all_indices)
        commit_losses = torch.stack(commit_losses)

        ret = (quantized, all_indices, commit_losses, *maybe_all_codes)
        return ret



from egogym.utils import make_cabinet_xml, make_drawer_xml
from egogym.components import Object, ArticulableObject

class ObjectsManager:
    def __init__(self, objects_dict, np_random, shuffle=True):
        self.np_random = np_random
        sorted_objects = sorted(objects_dict, key=lambda x: x[0])
        if shuffle:
            np_random.shuffle(sorted_objects)
        self.objects_dict = sorted_objects
        self.index = -1

    def sample(self, random=False):
        if random:
            index = self.np_random.integers(len(self.objects_dict))
        else:
            self.index = (self.index + 1) % len(self.objects_dict)
            index = self.index
        if self.objects_dict[index][1] == "procedurally_generated":
            return ArticulableObject(
                name=self.objects_dict[index][0],
                func=globals()[f"make_{self.objects_dict[index][0]}_xml"],
                np_random=self.np_random,
            )
        else:
            return Object(
                name=self.objects_dict[index][0],
                path=self.objects_dict[index][1],
                np_random=self.np_random,
            )



import random
from collections import OrderedDict
from datetime import timedelta

import accelerate
import numpy as np
import torch
import json

from . import *

# import accelerate


def set_seed_everywhere(seed):
    torch.manual_seed(seed)
    if torch.cuda.is_available():
        torch.cuda.manual_seed_all(seed)
    np.random.seed(seed)
    random.seed(seed)


def get_demos(cfg):
    train_files_list = cfg.dataset.train.config.trajectory_roots.trajectory_root_path
    val_files_list = cfg.dataset.test.config.trajectory_roots.trajectory_root_path

    train_demos = json.load(open(train_files_list))
    val_demos = json.load(open(val_files_list))

    return {"train": train_demos, "val": val_demos}


def setup_accelerate(experiment, num_epochs, task_title, save_every):
    ddp_kwargs = accelerate.DistributedDataParallelKwargs(find_unused_parameters=True)

    accelerator = accelerate.Accelerator(
        log_with=["wandb"],
        kwargs_handlers=[
            accelerate.InitProcessGroupKwargs(timeout=timedelta(hours=1.5)),
            ddp_kwargs,
        ],
    )
    accelerator.init_trackers(
        "experiment",
        config={
            "num_epochs": 1000,
            "task_title": task_title,
            "save_every": 20,
        },
    )

    return accelerator, ddp_kwargs


class AverageMeter:
    def __init__(self, value=None):
        super().__init__()
        if value is None:
            self._sum = 0
            self._count = 0
        else:
            self._sum = value
            self._count = 1

    def update(self, value, n=1):
        self._sum += value * n
        self._count += n
        self._avg = self._sum / self._count

    @property
    def avg(self):
        return self._avg


class AveragingDict(OrderedDict):
    def __init__(self, name="train", *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.name = name

    def update(self, other):
        for k, v in other.items():
            if k not in self:
                self[k] = AverageMeter(v)
            self[k].update(v)

    def __str__(self):
        return ", ".join(f"{self.name}/{k}: {v.avg:.2E}" for k, v in self.items())

    @property
    def full_summary(self):
        return {f"{self.name}/{k}": v.avg for k, v in self.items()}

    @property
    def summary(self):
        return {f"{self.name}/{k}": f"{v.avg:.2E}" for k, v in self.items()}


class Callbacks:
    """
    This class is used to implement callback functions inside training framework that
    passes, training params to callback functions inside model and loss objects.

    This is useful for custom logging inside models and losses like adding image
    visualisation, plotting hyperparameters like learning rate, etc.
    """

    def __init__(self):
        pass

    def set_workspace(self, workspace, accelerator):
        self.workspace = workspace
        self.accelerator = accelerator
        self._model_has_begin_epoch = hasattr(
            self.accelerator.unwrap_model(self.workspace.model), "_begin_epoch"
        )
        self._model_has_begin_batch = hasattr(
            self.accelerator.unwrap_model(self.workspace.model), "_begin_batch"
        )

        self._loss_fn_has_begin_epoch = hasattr(
            self.accelerator.unwrap_model(self.workspace.loss_fn), "_begin_epoch"
        )
        self._loss_fn_has_begin_batch = hasattr(
            self.accelerator.unwrap_model(self.workspace.loss_fn), "_begin_batch"
        )

    def begin_epoch(self):
        self.batch_step = 0

        if self._model_has_begin_epoch:
            model = self.accelerator.unwrap_model(self.workspace.model)
            is_main_process = self.accelerator.is_local_main_process
            begin_epoch_log = model._begin_epoch(
                epoch=self.workspace._epoch,
                epochs=self.workspace._cfg.num_epochs,
                train_dataloader=self.workspace._train_dataloaders,
                test_dataloader=self.workspace._test_dataloaders,
                optimizer=self.workspace.optimizer,
                scheduler=self.workspace.scheduler,
                is_main_process=is_main_process,
            )
            if begin_epoch_log is not None:
                self.accelerator.log(begin_epoch_log)

        if self._loss_fn_has_begin_epoch:
            loss_fn = self.accelerator.unwrap_model(self.workspace.loss_fn)
            is_main_process = self.accelerator.is_local_main_process
            begin_epoch_log = loss_fn._begin_epoch(
                epoch=self.workspace._epoch,
                epochs=self.workspace._cfg.num_epochs,
                train_dataloader=self.workspace._train_dataloaders,
                test_dataloader=self.workspace._test_dataloaders,
                optimizer=self.workspace.optimizer,
                scheduler=self.workspace.scheduler,
                is_main_process=is_main_process,
            )
            if begin_epoch_log is not None:
                self.accelerator.log(begin_epoch_log)

    def begin_batch(self):
        if self._model_has_begin_batch:
            model = self.accelerator.unwrap_model(self.workspace.model)
            is_main_process = self.accelerator.is_local_main_process
            begin_batch_log = model._begin_batch(
                epoch=self.workspace._epoch,
                batch_step=self.batch_step,
                epochs=self.workspace._cfg.num_epochs,
                train_dataloader=self.workspace._train_dataloaders,
                test_dataloader=self.workspace._test_dataloaders,
                optimizer=self.workspace.optimizer,
                scheduler=self.workspace.scheduler,
                is_main_process=is_main_process,
            )
            if begin_batch_log is not None:
                self.accelerator.log(begin_batch_log)

        if self._loss_fn_has_begin_batch:
            loss_fn = self.accelerator.unwrap_model(self.workspace.loss_fn)
            is_main_process = self.accelerator.is_local_main_process
            begin_batch_log = loss_fn._begin_batch(
                epoch=self.workspace._epoch,
                batch_step=self.batch_step,
                epochs=self.workspace._cfg.num_epochs,
                train_dataloader=self.workspace._train_dataloaders,
                test_dataloader=self.workspace._test_dataloaders,
                optimizer=self.workspace.optimizer,
                scheduler=self.workspace.scheduler,
                is_main_process=is_main_process,
            )
            if begin_batch_log is not None:
                self.accelerator.log(begin_batch_log)

        self.batch_step += 1


def log_wandb_image(image, label):
    """
    image: torch.Tensor or np.ndarray
    label: str, label for the image to categorize it in wandb

    returns: dict, {label: wandb.Image(image)}
    """
    import wandb

    if type(image) == torch.Tensor:
        image = image.cpu().numpy()
    while image.shape[0] == 1:
        image = image.squeeze(0)
    if image.shape[0] == 3:
        image = image.transpose(1, 2, 0)
    image = wandb.Image(image)
    return {label: image}



from egogym.embodiments.robot import Robot, requires_mj
from gymnasium.spaces import Box
import numpy as np


class Arm(Robot):
    
    def __init__(self, control_steps=2000, grasping_steps=1000, action_space="delta", np_random=None):
        super().__init__(control_steps=control_steps, grasping_steps=grasping_steps, action_space=action_space, np_random=np_random)
        self.joint_names = []
        self.qpos_ids = None
        self.act_ids = None
        self.gripper_act_id = None
    
    def additional_obs_spaces(self):
        num_joints = len(self.joint_names)
        return {
            "joint_positions": Box(low=-np.inf, high=np.inf, shape=(num_joints,), dtype=np.float32)
        }
    
    def action_space_shape(self):
        return (len(self.joint_names) + 1,)
    
    @requires_mj
    def get_additional_obs(self):
        return {
            "joint_positions": self.get_joints().astype(np.float32)
        }
    
    def setup_mj(self, model, data):
        super().setup_mj(model, data)
        if self.joint_names:
            self.qpos_ids = np.array([self.model.joint(name).qposadr[0] for name in self.joint_names])
            self.act_ids = np.array([self.model.actuator(name).id for name in self.joint_names])
        if hasattr(self, 'gripper_actuator_name'):
            self.gripper_act_id = self.model.actuator(self.gripper_actuator_name).id
    
    @requires_mj
    def get_joints(self):
        return self.data.qpos[self.qpos_ids].copy()
    
    @requires_mj
    def set_joints(self, joint_positions):
        self.data.ctrl[self.act_ids] = joint_positions
    
    @requires_mj
    def step(self, action, step_fn):

        if len(action.shape) == 2:
            action = action[0]
        
        num_joints = len(self.joint_names)
        desired_joints = action[:num_joints]
        grip_action = action[num_joints] if len(action) > num_joints else 0.0
        
        if self.action_space == "delta":
            target_joints = self.get_joints() + desired_joints
        else:
            target_joints = desired_joints
        

        self.set_joints(target_joints)
        self.data.ctrl[self.gripper_act_id] = 255.0 * np.clip(grip_action, 0.0, 1.0)
        
        step_fn(self.control_steps)



import logging
from enum import Enum
from typing import Dict, Optional, Tuple

import accelerate
import einops
import torch
import torch.nn as nn
import torch.nn.functional as F
import tqdm

from baselines.rum.models.bet.gpt import GPT
from baselines.rum.models.bet.utils import MLP


GENERATOR_SEED_FIXED = 123456789


class KMeansDiscretizer:
    """
    Simplified and modified version of KMeans algorithm from sklearn.
    We initialize this with a fixed seed to ensure that on each GPU we come up with the same
    clusters.
    """

    def __init__(
        self,
        num_bins: int = 100,
        kmeans_iters: int = 50,
    ):
        super().__init__()
        self.n_bins = num_bins
        self.kmeans_iters = kmeans_iters

    def fit(self, input_actions: torch.Tensor) -> None:
        self.bin_centers = KMeansDiscretizer._kmeans(
            input_actions, ncluster=self.n_bins, niter=self.kmeans_iters
        )

    @classmethod
    def _kmeans(cls, x: torch.Tensor, ncluster: int = 512, niter: int = 50):
        """
        Simple k-means clustering algorithm adapted from Karpathy's minGPT libary
        https://github.com/karpathy/minGPT/blob/master/play_image.ipynb
        """
        N, D = x.size()
        generator = torch.Generator()
        generator.manual_seed(GENERATOR_SEED_FIXED)

        c = x[
            torch.randperm(N, generator=generator)[:ncluster]
        ]  # init clusters at random, with a fixed seed

        pbar = tqdm.trange(niter)
        pbar.set_description("K-means clustering")
        for i in pbar:
            # assign all pixels to the closest codebook element
            a = ((x[:, None, :] - c[None, :, :]) ** 2).sum(-1).argmin(1)
            # move each codebook element to be the mean of the pixels that assigned to it
            c = torch.stack([x[a == k].mean(0) for k in range(ncluster)])
            # re-assign any poorly positioned codebook elements
            nanix = torch.any(torch.isnan(c), dim=1)
            ndead = nanix.sum().item()
            if ndead:
                tqdm.tqdm.write(
                    "done step %d/%d, re-initialized %d dead clusters"
                    % (i + 1, niter, ndead)
                )
            c[nanix] = x[
                torch.randperm(N, generator=generator)[:ndead]
            ]  # re-init dead clusters
        return c


class BehaviorTransformer(nn.Module):
    GOAL_SPEC = Enum("GOAL_SPEC", "concat stack unconditional")

    def __init__(
        self,
        obs_dim: int,
        act_dim: int,
        goal_dim: int,
        gpt_model: GPT,
        num_extra_predicted_actions: Optional[int] = None,
        trainable_obs_padding: bool = True,
        n_clusters: int = 32,
        kmeans_fit_steps: int = 500,
        kmeans_iters: int = 50,
        offset_loss_multiplier: float = 1.0e3,
        offset_distance_metric: str = "L2",
        gamma: float = 2.0,
        **kwargs,
    ):
        super().__init__()
        self._obs_dim = obs_dim
        self._act_dim = act_dim
        self._goal_dim = goal_dim
        self._num_extra_predicted_actions = num_extra_predicted_actions
        # Gradient-free, all zeros if we don't want to train this.
        self._obs_padding = nn.Parameter(
            trainable_obs_padding * torch.randn(obs_dim),
            requires_grad=trainable_obs_padding,
        )

        if goal_dim <= 0:
            self._cbet_method = self.GOAL_SPEC.unconditional
        elif obs_dim == goal_dim:
            self._cbet_method = self.GOAL_SPEC.concat
        else:
            self._cbet_method = self.GOAL_SPEC.stack

        self._gpt_model = gpt_model
        # For now, we assume the number of clusters is given.
        assert n_clusters > 0
        self._K = n_clusters
        self._kmeans_fit_steps = kmeans_fit_steps
        self._clustering_algo = KMeansDiscretizer(
            num_bins=n_clusters, kmeans_iters=kmeans_iters
        )
        self._current_steps = 0
        self._map_to_cbet_preds = MLP(
            in_channels=gpt_model.config.output_dim,
            hidden_channels=[(act_dim + 1) * n_clusters],
        )
        self._collected_actions = []
        self._have_fit_kmeans = False
        self._offset_loss_multiplier = offset_loss_multiplier
        # Placeholder for the cluster centers.
        generator = torch.Generator()
        generator.manual_seed(GENERATOR_SEED_FIXED)
        self.register_buffer(
            "_cluster_centers",
            torch.randn(
                (n_clusters, act_dim), generator=generator, dtype=torch.float32
            ),
        )
        self._criterion = FocalLoss(gamma=gamma, reduction="none")
        self._offset_criterion = (
            nn.MSELoss(reduction="none")
            if offset_distance_metric == "L2"
            else nn.L1Loss(reduction="none")
        )
        self._accelerator = accelerate.Accelerator()

    def _load_from_state_dict(self, *args, **kwargs):
        # Don't fit kmeans if we are loading from a state dict.
        self._current_steps = self._kmeans_fit_steps
        self._have_fit_kmeans = True
        return super()._load_from_state_dict(*args, **kwargs)

    def forward(
        self,
        obs_seq: torch.Tensor,
        goal_seq: Optional[torch.Tensor],
        action_seq: Optional[torch.Tensor],
        padding_seq: Optional[torch.Tensor],
        predict_with_offset: bool = True,
    ) -> Tuple[Optional[torch.Tensor], Optional[torch.Tensor]]:
        if self._current_steps == 0:
            self._cluster_centers = self._cluster_centers.to(obs_seq.device)
        if self._current_steps < self._kmeans_fit_steps and action_seq is not None:
            self._current_steps += 1
            self._fit_kmeans(obs_seq, goal_seq, action_seq, padding_seq)
        return self._predict(
            obs_seq,
            goal_seq,
            action_seq,
            padding_seq,
            predict_with_offset=predict_with_offset,
        )

    def _fit_kmeans(
        self,
        obs_seq: torch.Tensor,
        goal_seq: Optional[torch.Tensor],
        action_seq: Optional[torch.Tensor],
        padding_seq: Optional[torch.Tensor],
    ) -> Tuple[Optional[torch.Tensor], Optional[torch.Tensor]]:
        assert self._current_steps <= self._kmeans_fit_steps
        if self._current_steps == 1:
            self._cluster_centers = self._cluster_centers.to(action_seq.device)

        all_action_seq = self._accelerator.gather(action_seq)
        all_padding_seq = self._accelerator.gather(padding_seq)
        self._collected_actions.append(
            all_action_seq[torch.logical_not(all_padding_seq)]
        )
        if self._current_steps == self._kmeans_fit_steps:
            logging.info("Fitting KMeans")
            self._clustering_algo.fit(
                torch.cat(self._collected_actions, dim=0).view(-1, self._act_dim)
            )
            self._have_fit_kmeans = True
            self._cluster_centers = self._clustering_algo.bin_centers.float().to(
                action_seq.device
            )

    def _predict(
        self,
        obs_seq: torch.Tensor,
        goal_seq: Optional[torch.Tensor],
        action_seq: Optional[torch.Tensor],
        is_padded_action_seq: Optional[torch.Tensor],
        predict_with_offset: bool = True,
    ) -> Tuple[Optional[torch.Tensor], Optional[torch.Tensor], Dict[str, float]]:
        batch_size, obs_T, _ = obs_seq.shape
        _, action_T, _ = (
            action_seq.shape if action_seq is not None else (None, None, None)
        )
        # Take the one that is not None.
        actions_to_predict = action_T or obs_T
        if self._num_extra_predicted_actions:
            actions_to_predict += self._num_extra_predicted_actions
        # Now, figure out if we should pad the obs seq.
        if obs_T < actions_to_predict:
            # We need to pad the obs seq.
            pad_size = actions_to_predict - obs_T
            padded_obs_seq = torch.cat(
                [
                    obs_seq,
                    einops.repeat(
                        self._obs_padding, "D -> N T D", N=batch_size, T=pad_size
                    ),
                ],
                dim=1,
            )
        else:
            padded_obs_seq = obs_seq
        # Assume dimensions are N T D for N sequences of T timesteps with dimension D.
        if self._cbet_method == self.GOAL_SPEC.unconditional:
            gpt_input = padded_obs_seq
        elif self._cbet_method == self.GOAL_SPEC.concat:
            gpt_input = torch.cat([goal_seq, padded_obs_seq], dim=1)
        elif self._cbet_method == self.GOAL_SPEC.stack:
            gpt_input = torch.cat([goal_seq, padded_obs_seq], dim=-1)
        else:
            raise NotImplementedError

        gpt_output = self._gpt_model(gpt_input)
        if self._cbet_method == self.GOAL_SPEC.concat:
            # Chop off the goal encodings.
            gpt_output = gpt_output[:, goal_seq.size(1) :, :]
        cbet_preds = self._map_to_cbet_preds(gpt_output)
        cbet_logits, cbet_offsets = torch.split(
            cbet_preds, [self._K, self._K * self._act_dim], dim=-1
        )
        cbet_offsets = einops.rearrange(cbet_offsets, "N T (K A) -> N T K A", K=self._K)

        cbet_probs = torch.softmax(cbet_logits, dim=-1)
        N, T, choices = cbet_probs.shape
        # Sample from the multinomial distribution, one per row.
        sampled_centers = einops.rearrange(
            torch.multinomial(cbet_probs.view(-1, choices), num_samples=1),
            "(N T) 1 -> N T 1",
            N=N,
        )
        flattened_cbet_offsets = einops.rearrange(cbet_offsets, "N T K A -> (N T) K A")
        sampled_offsets = flattened_cbet_offsets[
            torch.arange(flattened_cbet_offsets.shape[0]), sampled_centers.flatten()
        ].view(N, T, self._act_dim)
        centers = self._cluster_centers[sampled_centers.flatten()].view(
            N, T, self._act_dim
        )
        a_hat = centers
        if predict_with_offset:
            a_hat += sampled_offsets
        if action_seq is None:
            return a_hat, None, {}
        # We are in training, so figure out the loss for the actions.
        # First, we need to find the closest cluster center for each action.
        action_bins = self._find_closest_cluster(action_seq)
        true_offsets = action_seq - self._cluster_centers[action_bins]
        predicted_offsets = flattened_cbet_offsets[
            torch.arange(flattened_cbet_offsets.shape[0]), action_bins.flatten()
        ].view(N, T, self._act_dim)
        # Now we can compute the loss.
        offset_loss = self._offset_criterion(predicted_offsets, true_offsets)
        cbet_loss = self._criterion(
            einops.rearrange(cbet_logits, "N T D -> (N T) D"),
            einops.rearrange(action_bins, "N T -> (N T)"),
        )
        # Now, use the padding mask to mask out the loss.
        if is_padded_action_seq is not None:
            cbet_loss *= ~is_padded_action_seq.view(-1)
            offset_loss *= ~is_padded_action_seq.unsqueeze(-1)
        cbet_loss, offset_loss = cbet_loss.mean(), offset_loss.mean()
        loss = cbet_loss + self._offset_loss_multiplier * offset_loss
        action_mse = F.mse_loss(a_hat, action_seq, reduction="none")
        action_l1 = F.l1_loss(a_hat, action_seq, reduction="none")
        norm = torch.norm(action_seq, p=2, dim=-1, keepdim=True) + 1e-9
        normalized_mse = (action_mse / norm).mean()
        if self._current_steps < self._kmeans_fit_steps:
            loss = loss.detach() + (loss * 0.0)

        loss_dict = {
            "classification_loss": cbet_loss.detach().cpu().item(),
            "offset_loss": offset_loss.detach().cpu().item(),
            "loss": loss.detach().cpu().item(),
            "L2_loss": action_mse.mean().detach().cpu().item(),
            "L2_loss_normalized": normalized_mse.mean().detach().cpu().item(),
            "L1_loss": action_l1.mean().detach().cpu().item(),
        }
        return a_hat, loss, loss_dict

    def _find_closest_cluster(self, action_seq: torch.Tensor) -> torch.Tensor:
        N, T, _ = action_seq.shape
        flattened_actions = einops.rearrange(action_seq, "N T A -> (N T) A")
        cluster_center_distance = torch.sum(
            (flattened_actions[:, None, :] - self._cluster_centers[None, :, :]) ** 2,
            dim=2,
        )  # (N T) K A -> (N T) K
        closest_cluster_center = torch.argmin(cluster_center_distance, dim=1)  # (N T)
        discretized_action = einops.rearrange(
            closest_cluster_center, "(N T) -> N T", N=N, T=T
        )
        return discretized_action

    def configure_optimizers(self, weight_decay, learning_rate, betas):
        optimizer = self._gpt_model.configure_optimizers(
            weight_decay=weight_decay,
            learning_rate=learning_rate,
            betas=betas,
        )
        optimizer.add_param_group({"params": self._map_to_cbet_preds.parameters()})
        return optimizer


class FocalLoss(nn.Module):
    def __init__(self, gamma: float = 0, reduction: str = "mean"):
        super(FocalLoss, self).__init__()
        self.gamma = gamma
        if reduction not in ("mean", "sum", "none"):
            raise NotImplementedError
        self.reduction = reduction

    def forward(self, input, target):
        logpt = F.log_softmax(input, dim=-1)
        logpt = logpt.gather(1, target.view(-1, 1)).view(-1)
        pt = logpt.exp()

        loss = -1 * (1 - pt) ** self.gamma * logpt
        if self.reduction == "mean":
            return loss.mean()
        elif self.reduction == "sum":
            return loss.sum()
        else:
            return loss



from egogym.wrappers.episode_monitor import EpisodeMonitor
from egogym.wrappers.unprivileged_molmo import UnprivilegedMolmo
from egogym.wrappers.unprivileged_gemini import UnprivilegedGemini
from egogym.wrappers.unprivileged_moondream import UnprivilegedMoondream
from egogym.wrappers.unprivileged_chatgpt import UnprivilegedChatGPT

VLM_WRAPPERS = {
    "molmo": UnprivilegedMolmo,
    "gemini": UnprivilegedGemini,
    "moondream": UnprivilegedMoondream,
    "chatgpt": UnprivilegedChatGPT,
}


def get_vlm_wrapper(vlm_name: str):
    if vlm_name not in VLM_WRAPPERS:
        available = ", ".join(f"'{k}'" for k in VLM_WRAPPERS.keys())
        raise ValueError(f"Unknown VLM option: '{vlm_name}'. Choose from {available}.")
    return VLM_WRAPPERS[vlm_name]




from scipy.spatial.transform import Rotation as R
import numpy as np

from egogym.utils import get_pose
from egogym.embodiments.robot import Robot, requires_mj


class FloatingGripper(Robot):    
    def __init__(self, control_steps=50, grasping_steps=1, action_space="delta", np_random=None):
        super().__init__(control_steps=control_steps, grasping_steps=grasping_steps, action_space=action_space, np_random=np_random)
        self.last_grasp = 0.0
    
    def action_space_shape(self):
        return (17,)

    def setup_mj(self, model, data):
        super().setup_mj(model, data)
        self.last_grasp = 0.0
        self.free_joint_id = model.joint("free_joint").qposadr[0]
    
    @requires_mj
    def prepare(self, desired_pose, step_fn=None):
        if self.add_noise_init:
            position_noise = self.np_random.uniform(-0.03, 0.03, size=3)
            desired_pose[:3, 3] += position_noise
            euler_angles = R.from_matrix(desired_pose[:3, :3]).as_euler("xyz")
            euler_angles[0] += self.np_random.uniform(-0.1, 0.1)
            euler_angles[1] += self.np_random.uniform(-0.05, 0.05)
            euler_angles[2] += self.np_random.uniform(-0.1, 0.1)
            desired_pose[:3, :3] = R.from_euler("xyz", euler_angles).as_matrix()
        self.last_teleop_pose = desired_pose
        self.data.mocap_pos[0] = desired_pose[:3, 3]
        self.data.mocap_quat[0] = np.roll(R.from_matrix(desired_pose[:3, :3]).as_quat(), 1)
        step_fn(self.control_steps)

    @requires_mj
    def move(self, desired_pose):
        self.data.mocap_pos[0] = desired_pose[:3, 3]
        self.data.mocap_quat[0] = np.roll(R.from_matrix(desired_pose[:3, :3]).as_quat(), 1)

    @requires_mj
    def step(self, action, step_fn):
        if len(action.shape) == 2:
            action = action[0]

        pose = action[:16].reshape(4,4)
        grip_action = action[16]
        
        if self.action_space == "delta":
            goal_pose = self.get_camera_pose() @ pose
        else:
            goal_pose = pose
        
        self.move(goal_pose)
        updated = self.set_grasp(grip_action)
        if updated:
            step_fn(self.grasping_steps)
        step_fn(self.control_steps)